# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import locale


class DeejaydError(Exception):
    """General purpose error structure."""

    # Handle unicode messages, what Exceptions cannot. See Python issue at
    # http://bugs.python.org/issue2517
    def __str__(self):
        if type(self.message) is unicode:
            return str(self.message.encode(locale.getpreferredencoding()))
        else:
            return str(self.message)

    def __unicode__(self):
        if type(self.message) is unicode:
            return self.message
        else:
            return unicode(self.message)

class DeejaydAnswer(object):
    """General purpose core answer container."""

    def __init__(self):
        self.contents = None
        self.error = False

    def set_error(self, msg):
        self.contents = msg
        self.error = True

    def get_contents(self):
        if self.error:
            raise DeejaydError(self.contents)
        return self.contents


class DeejaydKeyValue(DeejaydAnswer):
    """Dictionnary answer."""

    def __getitem__(self, name):
        self.get_contents()
        return self.contents[name]

    def keys(self):
        self.get_contents()
        return self.contents.keys()

    def items(self):
        self.get_contents()
        return self.contents.items()


class DeejaydList(DeejaydAnswer):
    """List answer."""

    def __len__(self):
        self.get_contents()
        return len(self.contents)

    def __iter__(self):
        self.get_contents()
        return self.contents.__iter__()


class DeejaydFileList(DeejaydAnswer):
    """File list answer."""

    def __init__(self):
        DeejaydAnswer.__init__(self)
        self.root_dir = ""
        self.files = []
        self.directories = []

    def set_rootdir(self, dir):
        self.root_dir = dir

    def add_file(self, file):
        self.files.append(file)

    def add_dir(self, dir):
        self.directories.append(dir)

    def set_files(self, files):
        self.files = files

    def set_directories(self, dirs):
        self.directories = dirs

    def get_files(self):
        self.get_contents()
        return self.files

    def get_directories(self):
        self.get_contents()
        return self.directories


class DeejaydMediaList(DeejaydAnswer):
    """Media list answer."""

    def __init__(self):
        DeejaydAnswer.__init__(self)
        self.medias = []
        self.filter = None
        self.sort = None
        self.media_type = None

    def add_media(self, media):
        self.medias.append(media)

    def get_medias(self):
        self.get_contents()
        return self.medias

    def set_medias(self, medias):
        self.medias = medias

    def is_magic(self):
        self.get_contents()
        return self.filter != None

    def set_media_type(self, media_type):
        self.media_type = media_type

    def get_media_type(self):
        self.get_contents()
        return self.media_type

    def set_filter(self, filter):
        self.filter = filter

    def get_filter(self):
        self.get_contents()
        return self.filter

    def set_sort(self, sort):
        self.sort = sort

    def get_sort(self):
        self.get_contents()
        return self.sort


class DeejaydDvdInfo(DeejaydAnswer):
    """Dvd information answer."""

    def __init__(self):
        DeejaydAnswer.__init__(self)
        self.dvd_content = {}

    def set_dvd_content(self, infos):
        self.dvd_content.update(infos)

    def add_track(self, track):
        if "track" not in self.dvd_content.keys():
            self.dvd_content['track'] = []
        self.dvd_content['track'].append(track)

    def set_tracks(self, tracks):
        self.dvd_content['track'] = tracks

    def get_dvd_contents(self):
        self.get_contents()
        if "track" not in self.dvd_content.keys():
            self.dvd_content['track'] = []
        return self.dvd_content


class DeejaydStaticPlaylist(object):
    """ Static playlist object """
    type = "static"

    def get(self, first=0, length=-1):
        raise NotImplementedError

    def add_path(self, path):
        return self.add_paths([path])

    def add_paths(self, paths):
        raise NotImplementedError

    def add_song(self, song_id):
        return self.add_songs([song_id])

    def add_songs(self, song_ids):
        raise NotImplementedError


class DeejaydMagicPlaylist(object):
    """ Magic playlist object """
    type = "magic"

    def get(self, first=0, length=-1):
        raise NotImplementedError

    def add_filter(self, filter):
        raise NotImplementedError

    def remove_filter(self, filter):
        raise NotImplementedError

    def clear_filters(self):
        raise NotImplementedError

    def get_properties(self):
        raise NotImplementedError

    def set_property(self, key, value):
        raise NotImplementedError


class DeejaydWebradioList(object):
    """Webradio list management."""

    def get(self, first = 0, length = None):
        raise NotImplementedError

    def get_available_sources(self):
        raise NotImplementedError

    def get_source_categories(self, source_name):
        raise NotImplementedError

    def set_source(self, source_name):
        raise NotImplementedError

    def set_source_categorie(self, categorie):
        raise NotImplementedError

    def add_webradio(self, name, urls):
        raise NotImplementedError

    def delete_webradio(self, wr_id):
        return self.delete_webradios([wr_id])

    def delete_webradios(self, wr_ids):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError


class DeejaydQueue(object):
    """Queue management."""

    def get(self, first = 0, length = None):
        raise NotImplementedError

    def add_path(self, path, pos = None):
        return self.add_paths([path], pos)

    def add_paths(self, paths, pos = None):
        raise NotImplementedError

    def add_song(self, song_id, pos = None):
        return self.add_songs([song_id], pos)

    def add_songs(self, song_ids, pos = None):
        raise NotImplementedError

    def load_playlist(self, pl_id, pos = None):
        return self.load_playlists([pl_id], pos)

    def load_playlists(self, pl_ids, pos = None):
        raise NotImplementedError

    def move(self, ids, new_pos):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError

    def del_song(self, id):
        return self.del_songs([id])

    def del_songs(self, ids):
        raise NotImplementedError


class DeejaydPanel(object):

    def get(self, first = 0, length = None):
        raise NotImplementedError

    def get_active_list(self):
        raise NotImplementedError

    def get_panel_tags(self):
        raise NotImplementedError

    def set_active_list(self, type, pl_id=""):
        raise NotImplementedError

    def set_panel_filters(self, tag, values):
        raise NotImplementedError

    def remove_panel_filters(self, tag):
        raise NotImplementedError

    def clear_panel_filters(self):
        raise NotImplementedError

    def set_search_filter(self, tag, value):
        raise NotImplementedError

    def clear_search_filter(self):
        raise NotImplementedError

    def set_sorts(self, sort):
        raise NotImplementedError


class DeejaydPlaylistMode(object):

    def get(self, first = 0, length = None):
        raise NotImplementedError

    def save(self, name):
        raise NotImplementedError

    def add_path(self, path, pos = None):
        return self.add_paths([path], pos)

    def add_paths(self, paths, pos = None):
        raise NotImplementedError

    def add_song(self, song_id, pos = None):
        return self.add_songs([song_id], pos)

    def add_songs(self, song_ids, pos = None):
        raise NotImplementedError

    def load(self, pl_id, pos = None):
        return self.loads([pl_id], pos)

    def loads(self, pl_ids, pos = None):
        raise NotImplementedError

    def move(self, ids, new_pos):
        raise NotImplementedError

    def shuffle(self):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError

    def del_song(self, id):
        return self.del_songs([id])

    def del_songs(self, ids):
        raise NotImplementedError


class DeejaydVideo(object):
    """Video management."""

    def get(self, first = 0, length = None):
        raise NotImplementedError

    def set(self, value, type = "directory"):
        raise NotImplementedError

    def set_sorts(self, sorts):
        raise NotImplementedError


class DeejaydSignal(object):

    SIGNALS = ('player.status',       # Player status change (play/pause/stop/
                                      # random/repeat/volume/manseek)
               'player.current',      # Currently played song change
               'player.plupdate',     # The current playlist has changed
               'playlist.listupdate', # The stored playlist list has changed
                                      # (either a saved playlist has been saved
                                      # or deleted).
               'playlist.update',     # A recorded playlist (static or magic)
                                      # has changed
                                      # set id of playlist as attribute
               'webradio.listupdate',
               'panel.update',
               'queue.update',
               'video.update',
               'dvd.update',
               'mode',                # Mode change
               'mediadb.aupdate',     # Media library audio update
               'mediadb.vupdate',     # Media library video update
               'mediadb.mupdate',     # a media has been updated
                                      # set id and type of media as attribute
                                      # set type of update as attribute
              )

    def __init__(self, name=None, attrs = {}):
        self.name = name
        self.attrs = attrs

    def set_name(self, name):
        self.name = name

    def get_name(self):
        return self.name

    def get_attrs(self):
        return self.attrs

    def get_attr(self, key):
        return self.attrs[key]

    def set_attr(self, key, value):
        self.attrs[key] = value


class DeejaydCore(object):
    """Abstract class for a deejayd core."""

    def __init__(self):
        self._clear_subscriptions()

    def ping(self):
        raise NotImplementedError

    def play_toggle(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError

    def previous(self):
        raise NotImplementedError

    def next(self):
        raise NotImplementedError

    def seek(self, pos, relative = False):
        raise NotImplementedError

    def get_current(self):
        raise NotImplementedError

    def go_to(self, id, id_type = None, source = None):
        raise NotImplementedError

    def set_volume(self, volume_value):
        raise NotImplementedError

    def set_option(self, source, option_name, option_value):
        raise NotImplementedError

    def set_mode(self, mode_name):
        raise NotImplementedError

    def get_mode(self):
        raise NotImplementedError

    def set_player_option(self, option_name, option_value):
        raise NotImplementedError

    def get_status(self):
        raise NotImplementedError

    def get_stats(self):
        raise NotImplementedError

    def update_audio_library(self, force = False, sync = False):
        raise NotImplementedError

    def update_video_library(self, force = False, sync = False):
        raise NotImplementedError

    def create_recorded_playlist(self, name, type):
        raise NotImplementedError

    def get_recorded_playlist(self, pl_id):
        raise NotImplementedError

    def erase_playlist(self, pl_ids):
        raise NotImplementedError

    def get_playlist_list(self):
        raise NotImplementedError

    def get_playlist(self):
        raise NotImplementedError

    def get_webradios(self):
        raise NotImplementedError

    def get_queue(self):
        raise NotImplementedError

    def get_panel(self):
        raise NotImplementedError

    def get_video(self):
        raise NotImplementedError

    def set_media_rating(self, media_ids, rating, type = "audio"):
        raise NotImplementedError

    def get_audio_dir(self, dir=None):
        raise NotImplementedError

    def get_audio_cover(self, media_id):
        raise NotImplementedError

    def audio_search(self, search_txt, type = 'all'):
        raise NotImplementedError

    def get_video_dir(self, dir=None):
        raise NotImplementedError

    def dvd_reload(self):
        raise NotImplementedError

    def mediadb_list(self, taglist, filter):
        raise NotImplementedError

    def get_dvd_content(self):
        raise NotImplementedError

    def __get_next_sub_id(self):
        sub_id = self.__sub_id_counter
        self.__sub_id_counter = self.__sub_id_counter + 1
        return sub_id

    def subscribe(self, signal_name, callback):
        """Subscribe to a signal with a callback. Returns an id."""
        if signal_name not in DeejaydSignal.SIGNALS:
            return DeejaydError('Unknown signal provided for subscription.')

        sub_id = self.__get_next_sub_id()
        self.__sig_subscriptions[sub_id] = (signal_name, callback)
        return sub_id

    def unsubscribe(self, sub_id):
        """Unsubscribe using the provied id."""
        try:
            del self.__sig_subscriptions[sub_id]
        except IndexError:
            raise DeejaydError('Unknown subscription id')

    def get_subscriptions(self):
        """Get the list of currently subcribed signals for this instance."""
        return dict([(sub_id, sub[0]) for (sub_id, sub)\
                                      in self.__sig_subscriptions.items()])

    def _clear_subscriptions(self):
        self.__sig_subscriptions = {}
        self.__sub_id_counter = 0

    def _dispatch_signal(self, signal):
        for cb in [sub[1] for sub in self.__sig_subscriptions.values()\
                                  if sub[0] == signal.get_name()]:
            cb(signal)

    def _dispatch_signame(self, signal_name, attrs = {}):
        self._dispatch_signal(DeejaydSignal(signal_name, attrs))


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os, re

from deejayd import __version__
from deejayd.interfaces import DeejaydError
from deejayd.mediafilters import *
from deejayd.rpc import Fault, INTERNAL_ERROR, INVALID_METHOD_PARAMS,\
                        DEEJAYD_PROTOCOL_VERSION
from deejayd.rpc.jsonbuilders import Get_json_filter
from deejayd.rpc.jsonparsers import Parse_json_filter
from deejayd.rpc.jsonrpc import JSONRPC, addIntrospection
from deejayd.rpc.rdfbuilder import modes as rdf_modes


def returns_answer(type, params = None):

    def returns_answer_instance(func):

        def returns_answer_func(*__args, **__kw):
            # verify arguments
            if params:
                for idx, p in enumerate(params):
                    try: arg = __args[idx+1]
                    except IndexError:
                        if p["req"] is True:
                            raise Fault(INVALID_METHOD_PARAMS,\
                                _("Param %s is required") % p["name"])
                        break
                    else:
                        if p['type'] == "int":
                            try: int(arg)
                            except (ValueError,TypeError):
                                raise Fault(INVALID_METHOD_PARAMS,\
                                    _("Param %s is not an int") % p["name"])
                        elif p['type'] in ("bool", "list", "dict"):
                            types = {"bool": bool, "dict": dict, "list": list}
                            if not isinstance(arg, types[p['type']]):
                                raise Fault(INVALID_METHOD_PARAMS,\
                                    _("Param %s has wrong type") % p["name"])
                        elif p['type'] == "int-list":
                            if not isinstance(arg, list):
                                raise Fault(INVALID_METHOD_PARAMS,\
                                    _("Param %s is not a list") % p["name"])
                            try: map(int, arg)
                            except (ValueError,TypeError):
                                raise Fault(INVALID_METHOD_PARAMS,\
                                  _("Param %s is not an int-list") % p["name"])
            try:
                res = func(*__args, **__kw)
            except DeejaydError, txt:
                raise Fault(INTERNAL_ERROR, str(txt))
            if type == "ack": res = True
            return {"type": type, "answer": res}

        returns_answer_func.__name__ = func.__name__
        returns_answer_func.__doc__ = func.__doc__
        returns_answer_func.answer_type = type
        returns_answer_func.params = params
        # build help
        p_desc = "\nNo arguments"
        if params:
            p_desc = "\nArguments : \n%s" %\
                "\n".join(["  * name '%s', type '%s', required '%s'" %\
                    (p["name"], p["type"], p["req"]) for p in params])
        returns_answer_func.help = """
Description :
    %(description)s
Answer type : %(answer)s %(p_desc)s""" % {\
                "description": func.__doc__,\
                "answer": type,\
                "p_desc": p_desc,\
            }
        # build signature
        p_dict = {
                "list": "array",
                "int-list": "array",
                "dict": "object",
                "string": "string",
                "int": "number",
                "bool": "boolean",
                "filter": "object",
                "sort": "array",
        }
        if params:
            signature = []
            opt_params = [p_dict[p["type"]] for p in params if not p["req"]]
            for i in range(len(opt_params)+1):
                s = ["object"]
                s.extend([p_dict[p["type"]] for p in params if p["req"]])
                s.extend(opt_params[:i])
                signature.append(s)
        else:
            signature = [["object"]]
        returns_answer_func.signature = signature

        return returns_answer_func

    return returns_answer_instance


class _DeejaydJSONRPC(JSONRPC):
    def __init__(self, deejayd = None):
        super(_DeejaydJSONRPC, self).__init__()
        self.deejayd_core = deejayd


class _DeejaydMainJSONRPC(_DeejaydJSONRPC):

    @returns_answer('ack')
    def jsonrpc_ping(self):
        """Does nothing, just replies with an acknowledgement that the
        command was received"""
        return None

    @returns_answer('ack', params=[{"name":"mode","type":"string","req":True}])
    def jsonrpc_setmode(self, mode):
        """Change the player mode. Possible values are :
  * playlist : to manage and listen songs in playlist
  * panel : to manage and listen songs in panel mode (like itunes)
  * video : to manage and wath video file
  * dvd : to wath dvd
  * webradio : to manage and listen webradios"""
        self.deejayd_core.set_mode(mode, objanswer=False)

    @returns_answer('dict')
    def jsonrpc_availablemodes(self):
        """For each available source, shows if it is activated or not.
   The answer consists in :
  * playlist : _bool_ true or false
  * panel : _bool_ true or false
  * webradio : _bool_ true or false (media backend has to be abble to read url streams)
  * video : _bool_ true or false (needs video dependencies, X display and needs to be activated in configuration)
  * dvd : _bool_ true or false (media backend has to be able to read dvd)"""
        return  dict(self.deejayd_core.get_mode(objanswer=False))

    @returns_answer('dict')
    def jsonrpc_status(self):
        """Return status of deejayd. Given informations are :
  * playlist : _int_ id of the current playlist
  * playlistlength : _int_ length of the current playlist
  * playlisttimelength : _int_ time length of the current playlist
  * playlistrepeat : _bool_ false (not activated) or true (activated)
  * playlistplayorder : inorder | random | onemedia | random-weighted
  * webradio : _int_ id of the current webradio list
  * webradiolength : _int_ number of recorded webradio
  * webradiosource : _str_ current source for webradio streams
  * webradiosourcecat : _str_ current categorie for webradio
  * queue : _int_ id of the current queue
  * queuelength : _int_ length of the current queue
  * queuetimelength : _int_ time length of the current queue
  * queueplayorder : _str_ inorder | random
  * video : _int_ id of the current video list
  * videolength : _int_ length of the current video list
  * videotimelength : _int_ time length of the current video list
  * videorepeat : _bool_ false (not activated) or true (activated)
  * videoplayorder : inorder | random | onemedia | random-weighted
  * dvd : _int_ id of the current dvd
  * dvdlength : _int_ number of tracks on the current dvd
  * volume : `[0-100]` current volume value
  * state : [play-pause-stop] the current state of the player
  * current : _int_:_int_:_str_ current media pos : current media file id : playing source name
  * time : _int_:_int_ position:length of the current media file
  * mode : [playlist-webradio-video] the current mode
  * audio_updating_db : _int_ show when a audio library update is in progress
  * audio_updating_error : _string_ error message that apppears when the audio library update has failed
  * video_updating_db : _int_ show when a video library update is in progress
  * video_updating_error : _string_ error message that apppears when the video library update has failed"""
        return dict(self.deejayd_core.get_status(objanswer = False))

    @returns_answer('dict')
    def jsonrpc_stats(self):
        """Return statistical informations :
  * audio_library_update : UNIX time of the last audio library update
  * video_library_update : UNIX time of the last video library update
  * videos : number of videos known by the database
  * songs : number of songs known by the database
  * artists : number of artists in the database
  * albums : number of albums in the database"""
        return dict(self.deejayd_core.get_stats(objanswer = False))

    @returns_answer('ack', [{"name":"source", "type":"string", "req":True},\
            {"name":"option_name", "type":"string","req":True},\
            {"name":"option_value","type":"string","req":True}])
    def jsonrpc_setOption(self, source, option_name, option_value):
        """Set player options "name" to "value" for mode "source", Available options are :
  * playorder (_str_: inorder, onemedia, random or random-weighted)
  * repeat (_bool_: True or False)"""
        self.deejayd_core.set_option(source, option_name,\
                option_value, objanswer=False)

    @returns_answer('ack', params=[\
            {"name":"ids", "type":"int-list", "req": True},\
            {"name": "value", "type": "int", "req": True},\
            {"name": "type", "type": "string", "req": False}])
    def jsonrpc_setRating(self, ids, value, type = "audio"):
        """Set rating of media file with ids equal to media_id for library 'type' """
        self.deejayd_core.set_media_rating(ids, value, type, objanswer=False)


class DeejaydTcpJSONRPC(_DeejaydMainJSONRPC):

    @returns_answer('ack')
    def jsonrpc_close(self):
        """Close the connection with the server"""
        return None

class DeejaydHttpJSONRPC(_DeejaydMainJSONRPC):

    @returns_answer('dict')
    def jsonrpc_serverInfo(self):
        """Return deejayd server informations :
  * server_version : deejayd server version
  * protocol_version : protocol version"""
        return {
            "server_version": __version__,
            "protocol_version": DEEJAYD_PROTOCOL_VERSION
        }

#
# Player commands
#
class DeejaydPlayerJSONRPC(_DeejaydJSONRPC):

    @returns_answer('ack')
    def jsonrpc_playToggle(self):
        """Toggle play/pause."""
        self.deejayd_core.play_toggle(objanswer=False)

    @returns_answer('ack')
    def jsonrpc_stop(self):
        """Stop playing."""
        self.deejayd_core.stop(objanswer=False)

    @returns_answer('ack')
    def jsonrpc_previous(self):
        """Go to previous song or webradio."""
        self.deejayd_core.previous(objanswer=False)

    @returns_answer('ack')
    def jsonrpc_next(self):
        """Go to next song or webradio."""
        self.deejayd_core.next(objanswer=False)

    @returns_answer('ack', params=[{"name":"pos", "type":"int", "req":True},\
            {"name":"relative", "type":"bool", "req":False}])
    def jsonrpc_seek(self, pos, relative = False):
        """Seeks to the position "pos" (in seconds) of the current media set relative argument to true to set new pos in relative way"""
        self.deejayd_core.seek(pos, relative, objanswer=False)

    @returns_answer('mediaList')
    def jsonrpc_current(self):
        """Return informations on the current song, webradio or video info."""
        medias = self.deejayd_core.get_current(objanswer=False)
        media_type = len(medias) == 1 and medias[0]["type"] or None
        return {
                "media_type": media_type,
                "medias": medias,
                "filter": None,
                "sort": None,
                }

    @returns_answer('ack',params=[\
            {"name":"id", "type":"string", "req":True},\
            {"name":"id_type","type":"string", "req":False},\
            {"name":"source","type":"string","req":False}])
    def jsonrpc_goto(self, id, id_type = "id", source = None):
        """Begin playing at media file with id "id" or toggle play/pause."""
        if not re.compile("^\w{1,}|\w{1,}\.\w{1,}$").search(id):
            raise Fault(INVALID_METHOD_PARAMS, _("Wrong id parameter"))
        self.deejayd_core.go_to(id, id_type, source, objanswer=False)

    @returns_answer('ack', params=[{"name":"volume", "type":"int", "req":True}])
    def jsonrpc_setVolume(self, volume):
        """Set volume to "volume". The volume range is 0-100."""
        self.deejayd_core.set_volume(volume, objanswer=False)

    @returns_answer('ack', params=[\
            {"name":"option_name", "type":"string", "req":True},\
            {"name":"option_value", "type":"string", "req":True}])
    def jsonrpc_setPlayerOption(self, option_name, option_value):
        """Set player option for the current media. Possible options are :
  * zoom : set zoom (video only), min=-85, max=400
  * audio_lang : select audio channel (video only)
  * sub_lang : select subtitle channel (video only)
  * av_offset : set audio/video offset (video only)
  * sub_offset : set subtitle/video offset (video only)
  * aspect_ratio : set video aspect ratio (video only), available values are :
    * auto
    * 1:1
    * 16:9
    * 4:3
    * 2.11:1 (for DVB)"""
        self.deejayd_core.set_player_option(option_name, option_value,\
                objanswer=False)


#
# media library
#
class _DeejaydLibraryJSONRPC(_DeejaydJSONRPC):
    type = ""

    @returns_answer('dict', params=[{"name":"force","type":"bool","req":False}])
    def jsonrpc_update(self, force = False):
        """Update the library.
  * 'type'_updating_db : the id of this task. It appears in the status until the update are completed."""
        func = getattr(self.deejayd_core, "update_%s_library"%self.type)
        return dict(func(force=force, objanswer=False))

    @returns_answer('fileAndDirList',\
            params=[{"name":"directory","type":"string","req":False}])
    def jsonrpc_getDir(self, directory = ""):
        """List the files of the directory supplied as argument."""
        func = getattr(self.deejayd_core, "get_%s_dir"%self.type)
        root_dir, dirs, files = func(directory, objanswer=False)
        type = self.type == "audio" and "song" or self.type
        return {
                "type": type,
                "files": files,
                "directories": dirs,
                "root": root_dir,
                }

    @returns_answer('mediaList',\
            params=[{"name":"pattern", "type":"string", "req":True},\
            {"name":"type","type":"string","req":True}])
    def jsonrpc_search(self, pattern, type):
        """Search files in library where "type" contains "pattern" content."""
        func = getattr(self.deejayd_core, "%s_search"%self.type)
        medias = func(pattern, type, objanswer=False)
        type = self.type == "audio" and "song" or self.type
        return {"media_type": type, "medias": medias}

class DeejaydAudioLibraryJSONRPC(_DeejaydLibraryJSONRPC):
    type = "audio"

    @returns_answer('list',\
            params=[{"name":"tag", "type":"string", "req":True},\
                {"name":"filter","type":"filter","req":False}])
    def jsonrpc_taglist(self, tag, filter = None):
        """List all the possible values for a tag according to the optional filter argument."""
        if filter is not None:
            filter = Parse_json_filter(filter)
        tag_list = self.deejayd_core.mediadb_list(tag, filter, objanswer=False)
        return list(tag_list)

class DeejaydVideoLibraryJSONRPC(_DeejaydLibraryJSONRPC):
    type = "video"

#
# generic class for modes
#
class _DeejaydModeJSONRPC(_DeejaydJSONRPC):
    def __init__(self, deejayd):
        super(_DeejaydModeJSONRPC, self).__init__(deejayd)
        self.source = getattr(self.deejayd_core, "get_%s"%self.source_name)()

    @returns_answer('mediaList',\
            params=[{"name":"first", "type":"int", "req":False},\
            {"name":"length","type":"int","req":False}])
    def jsonrpc_get(self, first=0, length=-1):
        """Return the content of this mode."""
        res = self.source.get(first, length, objanswer=False)
        if isinstance(res, tuple):
            medias, filter, sort = res
        else:
            medias, filter, sort = res, None, None
        json_filter = filter is not None and \
                Get_json_filter(filter).dump() or None
        return {
                "media_type": self.media_type,
                "medias": medias,
                "filter": json_filter,
                "sort": sort,
                }

#
# Dvd commands
#
class DeejaydDvdModeJSONRPC(_DeejaydJSONRPC):

    @returns_answer('dvdInfo')
    def jsonrpc_get(self):
        """Get the content of the current dvd."""
        return self.deejayd_core.get_dvd_content(objanswer=False)

    @returns_answer('ack')
    def jsonrpc_reload(self):
        """Load the content of the dvd player."""
        self.deejayd_core.dvd_reload(objanswer=False)

#
# video command
#
class DeejaydVideoModeJSONRPC(_DeejaydModeJSONRPC):
    media_type = "video"
    source_name = "video"

    @returns_answer('ack',\
            params=[{"name":"value", "type":"string", "req":True},\
            {"name":"type","type":"string","req":False}])
    def jsonrpc_set(self, value, type="directory"):
        """Set content of video mode"""
        self.source.set(value, type, objanswer=False)

    @returns_answer('ack', params=[{"name":"sort", "type":"sort", "req":True}])
    def jsonrpc_sort(self, sort):
        """Sort active medialist in video mode"""
        self.source.set_sorts(sort, objanswer=False)


#
# playlist mode command
#
class DeejaydPanelModeJSONRPC(_DeejaydModeJSONRPC):
    media_type = "song"
    source_name = "panel"

    @returns_answer('list')
    def jsonrpc_tags(self):
        """Return tag list used in panel mode."""
        tags = self.source.get_panel_tags(objanswer=False)
        return list(tags)

    @returns_answer('dict')
    def jsonrpc_activeList(self):
        """Return active list in panel mode
         * type : 'playlist' if playlist is choosen as active medialist 'panel' if panel navigation is active
         * value : if 'playlist' is selected, return used playlist id"""
        return dict(self.source.get_active_list(objanswer=False))

    @returns_answer('ack', [{"name":"type", "type":"string", "req":True},\
            {"name":"value", "type":"string", "req":False}])
    def jsonrpc_setActiveList(self, type, value = ""):
        """Set the active list in panel mode"""
        self.source.set_active_list(type, value, objanswer=False)

    @returns_answer('ack', [{"name":"tag", "type":"string", "req":True},\
            {"name":"values", "type":"list", "req":True}])
    def jsonrpc_setFilter(self, tag, values):
        """Set a filter for panel mode"""
        self.source.set_panel_filters(tag, values, objanswer=False)

    @returns_answer('ack', [{"name":"tag", "type":"string", "req":True},])
    def jsonrpc_removeFilter(self, tag):
        """Remove a filter for panel mode"""
        self.source.remove_panel_filters(tag, objanswer=False)

    @returns_answer('ack')
    def jsonrpc_clearFilter(self):
        """Clear filters for panel mode"""
        self.source.clear_panel_filters(objanswer=False)

    @returns_answer('ack', [{"name":"tag", "type":"string", "req":True},\
            {"name":"value", "type":"string", "req":True}])
    def jsonrpc_setSearch(self, tag, value):
        """Set search filter in panel mode"""
        self.source.set_search_filter(tag, value, objanswer=False)

    @returns_answer('ack')
    def jsonrpc_clearSearch(self):
        """Clear search filter in panel mode"""
        self.source.clear_search_filter(objanswer=False)

    @returns_answer('ack')
    def jsonrpc_clearAll(self):
        """Clear search filter and panel filters"""
        self.source.clear_search_filter(objanswer=False)
        self.source.clear_panel_filters(objanswer=False)

    @returns_answer('ack', [{"name":"sort","type":"sort","req":True},])
    def jsonrpc_setSort(self, sort):
        """Sort active medialist in panel mode"""
        self.source.set_sorts(sort, objanswer=False)
#
# playlist mode command
#
class DeejaydPlaylistModeJSONRPC(_DeejaydModeJSONRPC):
    media_type = "song"
    source_name = "playlist"

    @returns_answer('ack')
    def jsonrpc_clear(self):
        """Clear the current playlist."""
        self.source.clear(objanswer=False)

    @returns_answer('ack')
    def jsonrpc_shuffle(self):
        """Shuffle the current playlist."""
        self.source.shuffle(objanswer=False)

    @returns_answer('dict',\
            params=[{"name":"pls_name", "type":"string", "req":True}])
    def jsonrpc_save(self, pls_name):
        """Save the current playlist to "pls_name" in the database.
  * playlist_id : id of the recorded playlist"""
        return dict(self.source.save(pls_name, objanswer=False))

    @returns_answer('ack', params=[\
            {"name":"pl_ids", "type":"int-list", "req":True},
            {"name":"pos", "type":"int", "req":False}])
    def jsonrpc_loads(self, pl_ids, pos = None):
        """Load playlists passed as arguments "name" at the position "pos"."""
        self.source.loads(pl_ids, pos, objanswer=False)

    @returns_answer('ack', params=[\
            {"name":"paths", "type":"list", "req":True},
            {"name":"pos", "type":"int", "req":False}])
    def jsonrpc_addPath(self, paths, pos = None):
        """Load files or directories passed as arguments ("paths") at the position "pos" in the current playlist."""
        self.source.add_paths(paths, pos,objanswer=False)

    @returns_answer('ack', params=[\
            {"name":"ids", "type":"int-list", "req":True},
            {"name":"pos", "type":"int", "req":False}])
    def jsonrpc_addIds(self, ids, pos = None):
        """Load files with id passed as arguments ("ids") at the position "pos" in the current playlist."""
        self.source.add_songs(ids, pos, objanswer=False)

    @returns_answer('ack', params=[{"name":"ids","type":"int-list","req":True}])
    def jsonrpc_remove(self, ids):
        """Remove songs with ids passed as argument ("ids") from the current playlist"""
        self.source.del_songs(ids, objanswer=False)

    @returns_answer('ack', params=[\
            {"name":"ids", "type":"int-list", "req":True},
            {"name":"pos", "type":"int", "req":True}])
    def jsonrpc_move(self, ids, pos):
        """Move songs with id in "ids" to position "pos"."""
        self.source.move(ids, pos, objanswer=False)


#
# webradios commands
#
class DeejaydWebradioModeJSONRPC(_DeejaydModeJSONRPC):
    media_type = "webradio"
    source_name = "webradios"

    @returns_answer('dict')
    def jsonrpc_getAvailableSources(self):
        """Return list of available sources for webradio mode as source_name: has_categories"""
        return self.source.get_available_sources(objanswer=False)

    @returns_answer('list',\
                    params=[{"name":"source_name","type":"string","req":True}])
    def jsonrpc_getSourceCategories(self, source_name):
        """Return list of categories for webradio source 'source_name'"""
        return self.source.get_source_categories(source_name, objanswer=False)

    @returns_answer('ack', \
                    params=[{"name":"source_name","type":"string","req":True}])
    def jsonrpc_setSource(self, source_name):
        """Set current source to 'source_name'"""
        self.source.set_source(source_name, objanswer=False)

    @returns_answer('ack', \
                    params=[{"name":"categorie","type":"string","req":True}])
    def jsonrpc_setSourceCategorie(self, categorie):
        """Set categorie to 'categorie' for current source"""
        self.source.set_source_categorie(categorie, objanswer=False)

    @returns_answer('ack')
    def jsonrpc_localClear(self):
        """Remove all recorded webradios from the 'local' source."""
        self.source.clear(objanswer=False)

    @returns_answer('ack', params=[{"name":"ids","type":"int-list","req":True}])
    def jsonrpc_localRemove(self, ids):
        """Remove webradios with id in "ids" from the 'local' source."""
        self.source.delete_webradios(ids, objanswer=False)

    @returns_answer('ack', params=[{"name":"name","type":"string","req":True},\
            {"name":"url", "type":"list", "req":True}])
    def jsonrpc_localAdd(self, name, urls):
        """Add a webradio in 'local' source. Its name is "name" and the url of the webradio is "url". You can pass a playlist for "url" argument (.pls and .m3u format are supported)."""
        self.source.add_webradio(name, urls, objanswer=False)


#
# queue commands
#
class DeejaydQueueJSONRPC(_DeejaydModeJSONRPC):
    media_type = "song"
    source_name = "queue"

    @returns_answer('ack', params=[\
            {"name":"paths", "type":"list", "req":True},
            {"name":"pos", "type":"int", "req":False}])
    def jsonrpc_addPath(self, paths, pos = None):
        """Load files or directories passed as arguments ("paths") at the position "pos" in the queue."""
        self.source.add_paths(paths, pos,objanswer=False)

    @returns_answer('ack', params=[\
            {"name":"ids", "type":"int-list", "req":True},
            {"name":"pos", "type":"int", "req":False}])
    def jsonrpc_addIds(self, ids, pos = None):
        """Load files with id passed as arguments ("ids") at the position "pos" in the queue."""
        self.source.add_songs(ids, pos, objanswer=False)

    @returns_answer('ack', params=[\
            {"name":"pl_ids", "type":"int-list", "req":True},
            {"name":"pos", "type":"int", "req":False}])
    def jsonrpc_loads(self, pl_ids, pos = None):
        """Load playlists passed as arguments "name" at the position "pos"
        in the queue."""
        self.source.load_playlists(pl_ids, pos, objanswer=False)

    @returns_answer('ack', params=[{"name":"ids","type":"int-list","req":True}])
    def jsonrpc_remove(self, ids):
        """Remove songs with ids passed as argument ("ids") from the queue"""
        self.source.del_songs(ids, objanswer=False)

    @returns_answer('ack', params=[\
            {"name":"ids", "type":"int-list", "req":True},
            {"name":"pos", "type":"int", "req":True}])
    def jsonrpc_move(self, ids, pos):
        """Move songs with id in "ids" to position "pos"."""
        self.source.move(ids, pos, objanswer=False)

    @returns_answer('ack')
    def jsonrpc_clear(self):
        """Clear the queue."""
        self.source.clear(objanswer=False)


#
# recorded playlist
#
class DeejaydRecordedPlaylistJSONRPC(_DeejaydJSONRPC):

    @returns_answer('mediaList')
    def jsonrpc_list(self):
        """Return the list of recorded playlists."""
        pls = self.deejayd_core.get_playlist_list(objanswer=False)
        return {
                "media_type": 'playlist',
                "medias": pls,
                "filter": None,
                "sort": None,
                }

    @returns_answer('dict', [{"name":"name", "type":"string", "req":True},\
                    {"name":"type","type":"string","req":True}])
    def jsonrpc_create(self, name, type):
        """Create recorded playlist. The answer consist on
  * pl_id : id of the created playlist
  * name : name of the created playlist
  * type : type of the created playlist"""
        pl_infos = self.deejayd_core.create_recorded_playlist(\
                name, type, objanswer=False)
        return dict(pl_infos)

    @returns_answer('ack', [{"name":"pl_ids", "type":"int-list", "req":True}])
    def jsonrpc_erase(self, pl_ids):
        """Erase recorded playlists passed as arguments."""
        self.deejayd_core.erase_playlist(pl_ids, objanswer=False)

    @returns_answer('mediaList', [{"name":"pl_id", "type":"int", "req":True},\
                    {"name":"first", "type":"int", "req":False},\
                    {"name":"length","type":"int","req":False}])
    def jsonrpc_get(self, pl_id, first = 0, length = -1):
        """Return the content of a recorded playlist."""
        pls = self.deejayd_core.get_recorded_playlist(pl_id)
        if pls.type == "static":
            songs = pls.get(first, length, objanswer=False)
            filter, sort = None, None
        elif pls.type == "magic":
            songs, filter, sort = pls.get(first, length, objanswer=False)

        json_filter = filter is not None and \
                Get_json_filter(filter).dump() or None
        return {
                "media_type": 'song',
                "medias": songs,
                "filter": json_filter,
                "sort": sort,
                }

    @returns_answer('ack', params=[{"name":"pl_id", "type":"int", "req":True},\
                    {"name":"values", "type":"list", "req":True},\
                    {"name":"type","type":"string","req":False}])
    def jsonrpc_staticAdd(self, pl_id, values, type = "path"):
        """Add songs in a recorded static playlist. Argument 'type' has to be 'path' (default) or 'id'"""
        if type not in ("id", "path"):
            raise Fault(INVALID_METHOD_PARAMS,\
                _("Param 'type' has a wrong value"))
        pls = self.deejayd_core.get_recorded_playlist(pl_id)
        if pls.type == "magic":
            raise Fault(INVALID_METHOD_PARAMS,\
                _("Selected playlist is not static."))
        if type == "id":
            try: values = map(int, values)
            except (TypeError, ValueError):
                raise Fault(INVALID_METHOD_PARAMS,\
                        _("values arg must be integer"))
            pls.add_songs(values, objanswer=False)
        else:
            pls.add_paths(values, objanswer=False)

    @returns_answer('ack', params=[{"name":"pl_id", "type":"int", "req":True},\
                    {"name":"filter","type":"filter","req":True}])
    def jsonrpc_magicAddFilter(self, pl_id, filter):
        """Add a filter in recorded magic playlist."""
        pls = self.deejayd_core.get_recorded_playlist(pl_id)
        if pls.type != "magic":
            raise Fault(INVALID_METHOD_PARAMS,\
                _("Selected playlist is not magic."))
        if filter is not None:
            filter = Parse_json_filter(filter)
        pls.add_filter(filter, objanswer=False)

    @returns_answer('ack', params=[{"name":"pl_id", "type":"int", "req":True},\
                    {"name":"filter","type":"filter","req":True}])
    def jsonrpc_magicRemoveFilter(self, pl_id, filter):
        """Remove a filter from recorded magic playlist."""
        pls = self.deejayd_core.get_recorded_playlist(pl_id)
        if pls.type != "magic":
            raise Fault(INVALID_METHOD_PARAMS,\
                _("Selected playlist is not magic."))
        if filter is not None:
            filter = Parse_json_filter(filter)
        pls.remove_filter(filter, objanswer=False)

    @returns_answer('ack', params=[{"name":"pl_id", "type":"int", "req":True}])
    def jsonrpc_magicClearFilter(self, pl_id):
        """Remove all filter from recorded magic playlist."""
        pls = self.deejayd_core.get_recorded_playlist(pl_id)
        if pls.type != "magic":
            raise Fault(INVALID_METHOD_PARAMS,\
                _("Selected playlist is not magic."))
        pls.clear_filters(objanswer=False)

    @returns_answer('dict', [{"name":"pl_id", "type":"int", "req":True}])
    def jsonrpc_magicGetProperties(self, pl_id):
        """Get properties of a magic playlist
  * use-or-filter: if equal to 1, use "Or" filter instead of "And" (0 or 1)
  * use-limit: limit or not number of songs in the playlist (0 or 1)
  * limit-value: number of songs for this playlist (integer)
  * limit-sort-value: when limit is active sort playlist with this tag
  * limit-sort-direction: sort direction for limit (ascending or descending)"""
        pls = self.deejayd_core.get_recorded_playlist(pl_id)
        if pls.type != "magic":
            raise Fault(INVALID_METHOD_PARAMS,\
                _("Selected playlist is not magic."))
        return dict(pls.get_properties(objanswer=False))

    @returns_answer('ack', [{"name":"pl_id", "type":"int", "req":True},\
            {"name":"key", "type":"string","req":True},\
            {"name":"value", "type":"string", "req":True}])
    def jsonrpc_magicSetProperty(self, pl_id, key, value):
        """Set a property for a magic playlist."""
        pls = self.deejayd_core.get_recorded_playlist(pl_id)
        if pls.type != "magic":
            raise Fault(INVALID_METHOD_PARAMS,\
                _("Selected playlist is not magic."))
        pls.set_property(key, value, objanswer=False)


def build_protocol(deejayd, main):
    # add introspection
    addIntrospection(main)

    # add common deejayd subhandler
    sub_handlers = {
            "player": DeejaydPlayerJSONRPC,
            "audiolib": DeejaydAudioLibraryJSONRPC,
            "videolib": DeejaydVideoLibraryJSONRPC,
            "recpls": DeejaydRecordedPlaylistJSONRPC,
            }
    for key in sub_handlers:
        main.putSubHandler(key, sub_handlers[key](deejayd))
    # add mode deejayd subhandler
    mode_handlers = {
            "panel": DeejaydPanelModeJSONRPC,
            "webradio": DeejaydWebradioModeJSONRPC,
            "video": DeejaydVideoModeJSONRPC,
            "playlist": DeejaydPlaylistModeJSONRPC,
            "dvd": DeejaydDvdModeJSONRPC,
            "queue": DeejaydQueueJSONRPC,
            }
    for key in mode_handlers:
        try: mode = mode_handlers[key](deejayd)
        except DeejaydError: # this mode is not activated
            pass
        else:
            main.putSubHandler(key, mode)

    return main


#############################################################################
## part specific for jsonrpc over a TCP connecion : signals
#############################################################################
class DeejaydSignalJSONRPC(_DeejaydJSONRPC):

    def __init__(self, deejayd, connector):
        super(DeejaydSignalJSONRPC, self).__init__(deejayd)
        self.connector = connector

    @returns_answer('ack', [{"name":"signal", "type":"string", "req":True},\
            {"name":"value", "type":"bool", "req":True}])
    def jsonrpc_setSubscription(self, signal, value):
        """Set subscribtion to "signal" signal notifications to "value" which should be 0 or 1."""
        if value is False:
            self.connector.set_not_signaled(signal)
        elif value is True:
            self.connector.set_signaled(signal)

def set_signal_subhandler(deejayd, protocol):
    protocol.putSubHandler("signal", DeejaydSignalJSONRPC(deejayd, protocol))
    return protocol

#############################################################################
## part specific for jsonrpc over a http connecion
#############################################################################
class DeejaydWebJSONRPC(_DeejaydJSONRPC):
    def __init__(self, deejayd, tmp_dir):
        super(DeejaydWebJSONRPC, self).__init__(deejayd)
        self._tmp_dir = tmp_dir

    def __find_ids(self, pattern):
        ids = []
        for file in os.listdir(self._tmp_dir):
            if re.compile("^%s-[0-9]+" % pattern).search(file):
                t = file.split("-")[1] # id.ext
                t = t.split(".")
                try : ids.append(int(t[0]))
                except ValueError: pass
        return ids

    @returns_answer('dict', params=[{"name":"mid","type":"int","req":True}])
    def jsonrpc_writecover(self, mid):
        """ Record requested cover in the temp directory """
        try:
            cover = self.deejayd_core.get_audio_cover(mid,objanswer=False)
        except (TypeError, DeejaydError, KeyError):
            return {"cover": None}

        cover_ids = self.__find_ids("cover")
        ext = cover["mime"] == "image/jpeg" and "jpg" or "png"
        filename = "cover-%s.%s" % (str(cover["id"]), ext)
        if cover["id"] not in cover_ids:
            file_path = os.path.join(self._tmp_dir,filename)
            fd = open(file_path, "w")
            fd.write(cover["cover"])
            fd.close()
            os.chmod(file_path,0644)
            # erase unused cover files
            for id in cover_ids:
                try:
                    os.unlink(os.path.join(self._tmp_dir,\
                            "cover-%s.jpg" % id))
                    os.unlink(os.path.join(self._tmp_dir,\
                            "cover-%s.png" % id))
                except OSError:
                    pass
        return {"cover": os.path.join('tmp', filename), "mime": cover["mime"]}

    @returns_answer('dict', params=[{"name":"mode","type":"string","req":True}])
    def jsonrpc_buildSourceRDF(self, mode):
        """ Build rdf file with current medialist of the specified mode return dict with specific informations (like a description)"""
        try: rdf = rdf_modes[mode](self.deejayd_core, self._tmp_dir)
        except KeyError:
            raise Fault(INVALID_METHOD_PARAMS,_("mode %s is not known") % mode)
        return rdf.update()

    @returns_answer('dict',[{"name":"updated_tag","type":"string","req":False}])
    def jsonrpc_buildPanel(self, updated_tag = None):
        """ Build panel list """
        panel = self.deejayd_core.get_panel()

        medias, filters, sort = panel.get(objanswer=False)
        try: filter_list = filters.filterlist
        except (TypeError, AttributeError):
            filter_list = []

        answer = {"panels": {}}
        panel_filter = And()
        # find search filter
        for ft in filter_list:
            if ft.type == "basic" and ft.get_name() == "contains":
                panel_filter.combine(ft)
                answer["search"] = Get_json_filter(ft).dump()
                break
            elif ft.type == "complex" and ft.get_name() == "or":
                panel_filter.combine(ft)
                answer["search"] = Get_json_filter(ft).dump()
                break

        # find panel filter list
        for ft in filter_list:
            if ft.type == "complex" and ft.get_name() == "and":
                filter_list = ft
                break

        tag_list = panel.get_panel_tags(objanswer=False)
        try: idx = tag_list.index(updated_tag)
        except ValueError:
            pass
        else:
            tag_list = tag_list[idx+1:]

        for t in panel.get_panel_tags(objanswer=False):
            selected = []

            for ft in filter_list: # OR filter
                try: tag = ft[0].tag
                except (IndexError, TypeError): # bad filter
                    continue
                if tag == t:
                    selected = [t_ft.pattern for t_ft in ft]
                    tag_filter = ft
                    break

            if t in tag_list:
                list = self.deejayd_core.mediadb_list(t,\
                        panel_filter, objanswer=False)
                items = [{"name": _("All"), "value":"__all__", \
                    "class":"list-all", "sel":str(selected==[]).lower()}]
                if t == "various_artist" and "__various__" in list:
                    items.append({"name": _("Various Artist"),\
                        "value":"__various__",\
                        "class":"list-unknown",\
                        "sel":str("__various__" in selected).lower()})
                items.extend([{"name": l,"value":l,\
                    "sel":str(l in selected).lower(), "class":""}\
                    for l in list if l != "" and l != "__various__"])
                if "" in list:
                    items.append({"name": _("Unknown"), "value":"",\
                        "class":"list-unknown",\
                        "sel":str("" in selected).lower()})
                answer["panels"][t] = items
            # add filter for next panel
            if len(selected) > 0:
                panel_filter.combine(tag_filter)

        return answer

def set_web_subhandler(deejayd, tmp_dir, main):
    main.putSubHandler("web", DeejaydWebJSONRPC(deejayd, tmp_dir))
    return main

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

class Table(object):
    """Declare a table in a database schema."""

    def __init__(self, name, key=[]):
        self.name = name
        self.columns = []
        self.indices = []
        self.key = key
        if isinstance(key, basestring):
            self.key = [key]

    def __getitem__(self, objs):
        self.columns = [o for o in objs if isinstance(o, Column)]
        self.indices = [o for o in objs if isinstance(o, Index)]
        return self


class Column(object):
    """Declare a table column in a database schema."""

    def __init__(self, name, type='text', size=None, auto_increment=False):
        self.name = name
        self.type = type
        self.size = size
        self.auto_increment = auto_increment


class Index(object):
    """Declare an index for a database schema."""

    def __init__(self, columns, unique = True):
        self.unique = unique
        self.columns = columns


db_schema_version=13
db_schema = [
    Table('library_dir', key='id')[
        Column('id', auto_increment=True),
        Column('name'),
        Column('lib_type'), # audio or video
        Column('type'), # directory or dirlink
        Index(('name', 'lib_type', 'type'))],
    Table('library', key='id')[
        Column('id', auto_increment=True),
        Column('directory', type='int'),
        Column('name'),
        Column('lastmodified', type='int'),
        Index(('name', 'directory')),
        Index(('directory',), unique = False)],
    Table('media_info', key=('id','ikey'))[
        Column('id', type="int"),
        Column('ikey'),
        Column('value')],
    Table('cover', key='id')[
        Column('id', auto_increment=True),
        # path to the cover file or
        # hash of the picture for cover inside audio file
        Column('source'),
        Column('mime_type'), # mime type of the cover, ex image/jpeg
        Column('lmod', type="int"), # last modified
        Column('image', type="blob"),
        Index(('source',))],
    Table('medialist', key='id')[
        Column('id', auto_increment=True),
        Column('name'),
        Column('type'), # magic or static
        Index(('name','type'))],
    Table('medialist_libraryitem', key=('position'))[
        Column('position', auto_increment=True),
        Column('medialist_id', type='int'),
        Column('libraryitem_id', type='int')],
    Table('medialist_property', key=('medialist_id','ikey'))[
        Column('medialist_id', type='int'),
        Column('ikey'),
        Column('value')],
    Table('medialist_sorts', key=('position'))[
        Column('position', auto_increment=True),
        Column('medialist_id', type='int'),
        Column('tag'),
        Column('direction')],
    Table('medialist_filters', key=('medialist_id', 'filter_id'))[
        Column('medialist_id', type='int'),
        Column('filter_id', type='int')],
    Table('filters', key=('filter_id'))[
        Column('filter_id', auto_increment=True),
        Column('type')], # complex or basic
    Table('filters_basicfilters', key=('filter_id'))[
        Column('filter_id', type='int'),
        Column('tag'), # criterion
        Column('operator'), # equal, not equal, regex, regexi, etc.
        Column('pattern')], # matched value
    Table('filters_complexfilters', key=('filter_id'))[
        Column('filter_id', type='int'),
        Column('combinator')], # AND, OR, XOR
    Table('filters_complexfilters_subfilters',
          key=('complexfilter_id', 'filter_id'))[
        Column('complexfilter_id', type='int'),
        Column('filter_id', type='int')],
    Table('webradio', key='id')[
        Column('id', auto_increment=True),
        Column('name'),
        Index(('name',), unique = False)],
    Table('webradio_entries', key='id')[
        Column('id', auto_increment=True),
        Column('url'),
        Column('webradio_id', type='int'),
        Index(('url', 'webradio_id'))],
    Table('stats', key='name')[
        Column('name'),
        Column('value', type='int')],
    Table('variables', key='name')[
        Column('name'),
        Column('value')],
    ]
db_init_cmds = [
    # stats
    "INSERT INTO stats VALUES('video_library_update',0);",
    "INSERT INTO stats VALUES('audio_library_update',0);",
    "INSERT INTO stats VALUES('songs',0);",
    "INSERT INTO stats VALUES('videos',0);",
    "INSERT INTO stats VALUES('artists',0);",
    "INSERT INTO stats VALUES('albums',0);",
    "INSERT INTO stats VALUES('genres',0);",
    # variables
    "INSERT INTO variables VALUES('volume','0');",
    "INSERT INTO variables VALUES('current','-1');",
    "INSERT INTO variables VALUES('current_source','none');",
    "INSERT INTO variables VALUES('current_pos','0');",
    "INSERT INTO variables VALUES('state','stop');",
    "INSERT INTO variables VALUES('source','playlist');",
    "INSERT INTO variables VALUES('database_version','13');",
    ]

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import sys, os, shutil
import unittest

from testdeejayd.databuilder import TestData, TestAudioCollection,\
                                    TestVideoCollection
from testdeejayd.server import TestServer
from deejayd.ui.config import DeejaydConfig


class DeejaydTest(unittest.TestCase):

    def setUp(self):
        from deejayd.ui.i18n import DeejaydTranslations
        t = DeejaydTranslations()
        t.install()


class TestCaseWithData(DeejaydTest):

    def setUp(self):
        super(TestCaseWithData, self).setUp()
        self.testdata = TestData()

    def assert_filter_matches_sample(self, retrieved_filter):
        self.assertEqual(retrieved_filter.__class__.__name__, 'And')
        anded = retrieved_filter.filterlist
        self.assertEqual(anded[0].__class__.__name__, 'Contains')
        self.assertEqual(anded[0].tag, 'artist')
        self.assertEqual(anded[0].pattern, 'Britney')
        self.assertEqual(anded[1].__class__.__name__, 'Or')
        ored = anded[1].filterlist
        self.assertEqual(ored[0].__class__.__name__, 'Equals')
        self.assertEqual(ored[0].tag, 'genre')
        self.assertEqual(ored[0].pattern, 'Classical')
        self.assertEqual(ored[1].__class__.__name__, 'Equals')
        self.assertEqual(ored[1].tag, 'genre')
        self.assertEqual(ored[1].pattern, 'Disco')


class _TestCaseWithMediaData(DeejaydTest):

    def setUp(self):
        super(_TestCaseWithMediaData, self).setUp()
        self.testdata = self.collection_class()
        self.testdata.buildLibraryDirectoryTree()

    def tearDown(self):
        self.testdata.cleanLibraryDirectoryTree()


class TestCaseWithAudioData(_TestCaseWithMediaData):
    collection_class = TestAudioCollection


class TestCaseWithVideoData(_TestCaseWithMediaData):
    collection_class = TestVideoCollection


class TestCaseWithAudioAndVideoData(DeejaydTest):
    video_support = True

    def setUp(self):
        super(TestCaseWithAudioAndVideoData, self).setUp()

        self.testdata = TestData()
        # audio library
        self.test_audiodata = TestAudioCollection()
        self.test_audiodata.buildLibraryDirectoryTree()

        # video library
        self.test_videodata = TestVideoCollection()
        self.test_videodata.buildLibraryDirectoryTree()

    def tearDown(self):
        self.test_audiodata.cleanLibraryDirectoryTree()
        self.test_videodata.cleanLibraryDirectoryTree()


class TestCaseWithServer(TestCaseWithAudioAndVideoData):
    profiles = "default"

    def setUp(self):
        super(TestCaseWithServer, self).setUp()
        # create custom configuration files
        current_dir = os.path.dirname(__file__)
        DeejaydConfig.custom_conf = os.path.join(current_dir,\
                "profiles", self.profiles)
        config = DeejaydConfig()
        config.set('mediadb', 'music_directory',\
                self.test_audiodata.getRootDir())
        config.set('mediadb', 'video_directory',\
                self.test_videodata.getRootDir())
        self.dbfilename = None
        if config.get('database', 'db_type') == 'sqlite':
            self.dbfilename = '/tmp/testdeejayddb-' +\
                    self.testdata.getRandomString() + '.db'
            config.set('database', 'db_name', self.dbfilename)
        elif config.get('database', 'db_type') == 'mysql':
            import MySQLdb
            try:
                connection = MySQLdb.connect(\
                    db=config.get('database', 'db_name'),\
                    user=config.get('database', 'db_user'),\
                    passwd=config.get('database', 'db_password'),\
                    host=config.get('database', 'db_host'),\
                    port=config.getint('database', 'db_port'),\
                    use_unicode=True, charset="utf8")
            except MySQLdb.DatabaseError:
                print "Unable to connect to mysql db"
                sys.exit(1)
            cursor = connection.cursor()
           # drop all table
            from deejayd.database.schema import db_schema
            for table in db_schema:
                try:
                    cursor.execute("DROP TABLE `%s`" % table.name)
                except MySQLdb.DatabaseError:
                    pass
           # commit changes and close
            connection.commit()
            connection.close()

        self.tmp_dir = None
        if config.getboolean("webui","enabled"):
            # define a tmp directory
            self.tmp_dir = '/tmp/testdeejayd-tmpdir-'+\
                    self.testdata.getRandomString()
            config.set("webui", "tmp_dir", self.tmp_dir)

        # record config to be used by the test server
        self.conf = os.path.join(current_dir, "profiles", "current")
        fp = open(self.conf, "w")
        config.write(fp)
        fp.close()
        os.chmod(self.conf, 0644)

        # launch server
        self.testserver = TestServer(self.conf)
        self.testserver.start()

        # record port for clients
        self.serverPort = config.getint('net', 'port')
        self.webServerPort = config.getint('webui', 'port')

        # update video_support var
        self.video_support = config.get("general","media_backend")!="gstreamer"

    def tearDown(self):
        self.testserver.stop()
        if self.dbfilename is not None: # Clean up temporary db file
            os.unlink(self.dbfilename)
        if self.tmp_dir is not None:
             try: shutil.rmtree(self.tmp_dir)
             except (IOError, OSError):
                 pass
        os.unlink(self.conf)
        super(TestCaseWithServer, self).tearDown()

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Deejayd DB testing module
"""
import os,time
from testdeejayd import TestCaseWithAudioData, TestCaseWithVideoData
from deejayd import database, mediafilters, plugins
from deejayd.database.queries import DatabaseQueries
from deejayd.mediadb.library import AudioLibrary,VideoLibrary,NotFoundException
from deejayd.mediadb import inotify

# FIXME : Those imports should really go away one day
from deejayd.player import xine
from deejayd.ui.config import DeejaydConfig


class _TestDeejayDBLibrary(object):

    def setUp(self):
        self.dbfilename = '/tmp/testdeejayddb-'+self.testdata.getRandomString()

        # init player
        config = DeejaydConfig()
        config.set('database', 'db_type', 'sqlite')
        config.set('database', 'db_name', self.dbfilename)
        config.set('xine','audio_output',"none")
        config.set('xine','video_output',"none")

        self.db = database.init(config)
        player = xine.XinePlayer(self.db, plugins.PluginManager(config), config)
        player.init_video_support()

        self.library = self.__class__.library_class(self.db, player, \
                                                    self.testdata.getRootDir())
        self.library._update()

        self.do_update = True

    def tearDown(self):
        self.db.close()
        os.remove(self.dbfilename)

    def verifyMediaDBContent(self, testTag=True):
        time.sleep(0.2)
        if self.do_update: # First update mediadb
            self.library._update()
        else:
            time.sleep(1.5)

        self.assertRaises(NotFoundException,
                  self.library.get_dir_content, self.testdata.getRandomString())
        self.assertRaises(NotFoundException,
                  self.library.get_file, self.testdata.getRandomString())

        self.__verifyRoot(self.testdata.getRootDir(), testTag)

    def __verifyRoot(self, requested_root, testTag=True, inlink_path=None):
        for root, dirs, files in os.walk(requested_root):
            try: root = root.decode("utf-8", "strict").encode("utf-8")
            except UnicodeError:
                continue

            strip_root = self.testdata.stripRoot(root)

            new_dirs = []
            for d in dirs:
                try: d = d.decode("utf-8", "strict").encode("utf-8")
                except UnicodeError:
                    continue
                dir_path = os.path.join(root, d)
                if os.path.islink(dir_path):
                    rel_path = os.path.join(strip_root, d)
                    self.__verifyRoot(dir_path, testTag, rel_path)
                new_dirs.append(d)
            dirs = new_dirs

            new_files = []
            for f in files:
                try: f = f.decode("utf-8", "strict").encode("utf-8")
                except UnicodeError:
                    continue
                if f != "cover.jpg" and not f.endswith(".srt"):
                    new_files.append(f)
            files = new_files

            try: contents = self.library.get_dir_content(strip_root)
            except NotFoundException:
                allContents = self.library.get_dir_content('')
                self.assert_(False,
                    "'%s' is in directory tree but was not found in DB %s" %\
                    (root,str(allContents)))

            # First, verify directory list
            self.assertEqual(len(contents["dirs"]), len(dirs))
            for dir in dirs:
                self.assert_(dir in contents["dirs"],
                    "'%s' is in directory tree but was not found in DB %s in current root '%s'" % (dir,str(contents["dirs"]),root))

            # then, verify file list
            self.assertEqual(len(contents["files"]), len(files))
            db_files = [f["filename"] for f in contents["files"]]
            for file in files:
                (name,ext) = os.path.splitext(file)
                if ext.lower() in self.__class__.supported_ext:
                    self.assert_(file in db_files,
                    "'%s' is a file in directory tree but was not found in DB"\
                    % file)
                    relPath = os.path.join(strip_root, file)
                    if testTag:
                        self.verifyTag(relPath, inlink_path)

    def verifyTag(self, filePath, inlink_path=None):
        try: inDBfile = self.library.get_file(filePath)
        except NotFoundException:
            self.assert_(False,
                "'%s' is a file in directory tree but was not found in DB"\
                % filePath)
        else: inDBfile = inDBfile[0]

        if inlink_path:
            link_full_path = os.path.join(self.testdata.getRootDir(),
                                          inlink_path)
            abs_path = filePath[len(inlink_path)+1:]
            realFile = self.testdata.dirlinks[link_full_path].medias[abs_path]
        else:
            realFile = self.testdata.medias[filePath]
        return (inDBfile, realFile)

        for tag in ("title","artist","album","genre"):
            self.assert_(realFile[tag] == inDBfile[tag],
                "tag %s for %s different between DB and reality %s != %s" % \
                (tag,realFile["filename"],realFile[tag],inDBfile[tag]))

    def testDirlinks(self):
        self.testdata.addDirLink()
        self.verifyMediaDBContent()

        self.testdata.moveDirLink()
        self.verifyMediaDBContent()

        self.testdata.removeDirLink()
        self.verifyMediaDBContent()


class TestVideoLibrary(TestCaseWithVideoData, _TestDeejayDBLibrary):
    library_class = VideoLibrary
    supported_ext = (".mpg",".avi")

    def setUp(self):
        TestCaseWithVideoData.setUp(self)
        _TestDeejayDBLibrary.setUp(self)

    def tearDown(self):
        _TestDeejayDBLibrary.tearDown(self)
        TestCaseWithVideoData.tearDown(self)

    def testGetDir(self):
        """built directory detected by video library"""
        self.verifyMediaDBContent()

    def testAddSubdirectory(self):
        """Add a subdirectory in video library"""
        self.testdata.addSubdir()
        self.verifyMediaDBContent()

    def testAddMedia(self):
        """Add a media file in video library"""
        self.testdata.addMedia()
        self.verifyMediaDBContent()

    def testAddSubtitle(self):
        """Add a subtitle file in video library"""
        self.testdata.add_subtitle()
        self.verifyMediaDBContent()

    def testRemoveSubtitle(self):
        """Remove a subtitle file in video library"""
        self.testdata.remove_subtitle()
        self.verifyMediaDBContent()

    def verifyTag(self, filePath, inlink_path=None):
        (inDBfile, realFile) = _TestDeejayDBLibrary.verifyTag(self,
                                                         filePath, inlink_path)
        for tag in ('length','videowidth','videoheight','external_subtitle'):
            self.assertEqual(str(realFile[tag]), str(inDBfile[tag]))


class TestAudioLibrary(TestCaseWithAudioData, _TestDeejayDBLibrary):
    library_class = AudioLibrary
    supported_ext = (".ogg",".mp3",".mp4",".flac")

    def setUp(self):
        TestCaseWithAudioData.setUp(self)
        _TestDeejayDBLibrary.setUp(self)

    def tearDown(self):
        TestCaseWithAudioData.tearDown(self)
        _TestDeejayDBLibrary.tearDown(self)

    def testGetDir(self):
        """built directory detected by audio library"""
        self.verifyMediaDBContent()

    def testAddDirectory(self):
        """Add a directory in audio library"""
        self.testdata.addDir()
        self.verifyMediaDBContent()

    def testAddSubdirectory(self):
        """Add a subdirectory in audio library"""
        self.testdata.addSubdir()
        self.verifyMediaDBContent()

    def testAddSubSubdirectory(self):
        """Add a subsubdirectory in audio library"""
        self.testdata.addSubSubdir()
        self.verifyMediaDBContent()

    def testRenameDirectory(self):
        """Rename a directory in audio library"""
        self.testdata.renameDir()
        self.verifyMediaDBContent(False)

    def testRemoveDirectory(self):
        """Remove a directory in audio library"""
        self.testdata.removeDir()
        self.verifyMediaDBContent()

    def testAddMedia(self):
        """Add a media file in audio library"""
        self.testdata.addMedia()
        self.verifyMediaDBContent()

    def testRenameMedia(self):
        """Rename a media file in audio library"""
        self.testdata.renameMedia()
        self.verifyMediaDBContent(False)

    def testRemoveMedia(self):
        """Remove a media file in audio library"""
        self.testdata.removeMedia()
        self.verifyMediaDBContent()

    def testChangeTag(self):
        """Tag value change detected by audio library"""
        self.testdata.changeMediaTags()
        self.verifyMediaDBContent()

    def testAddCover(self):
        """Add cover in audio library"""
        self.testdata.add_cover()
        self.verifyMediaDBContent()

    def testRemoveCover(self):
        """Add cover in audio library"""
        self.testdata.remove_cover()
        self.verifyMediaDBContent()

    def testSearchFile(self):
        """Search a file in audio library"""
        filter = mediafilters.Contains("genre", self.testdata.getRandomString())
        self.assertEqual([], self.library.search(filter))

        searched_genre = self.testdata.getRandomGenre()
        filter = mediafilters.Contains("genre", searched_genre)
        matched_medias_uri = []
        for path, media in self.testdata.medias.items():
            if searched_genre == media.tags['genre']:
                matched_medias_uri.append(media.tags['uri'])

        found_items_uri = [x['uri'] for x in self.library.search(filter)]

        self.assertEqual(len(matched_medias_uri), len(found_items_uri))

        found_items_uri.sort()
        matched_medias_uri.sort()
        self.assertEqual(found_items_uri, matched_medias_uri)

    def verifyTag(self,filePath, inlink_path=None):
        (inDBfile, realFile) = _TestDeejayDBLibrary.verifyTag(self, filePath,
                                                              inlink_path)
        for tag in ("title","artist","album","genre"):
            self.assert_(realFile[tag] == inDBfile[tag],
                "tag %s for %s different between DB and reality %s != %s" % \
                (tag,realFile["filename"],realFile[tag],inDBfile[tag]))


class TestInotifySupport(TestCaseWithAudioData, _TestDeejayDBLibrary):
    library_class = AudioLibrary
    supported_ext = (".ogg",".mp3",".mp4",".flac")

    def setUp(self):
        TestCaseWithAudioData.setUp(self)
        _TestDeejayDBLibrary.setUp(self)

        self.do_update = False

        # start inotify thread
        self.watcher = inotify.get_watcher(self.db, self.library, None)
        self.watcher.start()
        time.sleep(5)

    def tearDown(self):
        self.watcher.close()

        _TestDeejayDBLibrary.tearDown(self)
        TestCaseWithAudioData.tearDown(self)

    def testAddMedia(self):
        """Inotify support : Add a media in audio library"""
        self.testdata.addMedia()
        self.verifyMediaDBContent()

    def testAddSubdirectory(self):
        """Inotify support : Add a subdirectory"""
        self.testdata.addSubdir()
        self.verifyMediaDBContent()

    def testAddSubSubdirectory(self):
        """Inotify support : Add a subsubdirectory"""
        self.testdata.addSubSubdir()
        self.verifyMediaDBContent()

    def testRenameDirectory(self):
        """Inotify support : Rename a directory"""
        self.testdata.renameDir()
        self.verifyMediaDBContent(False)

    def testRemoveDirectory(self):
        """Inotify support : Remove a directory"""
        self.testdata.removeDir()
        self.verifyMediaDBContent()

    def testChangeTag(self):
        """Inotify support : Tag value change detected"""
        self.testdata.changeMediaTags()
        self.verifyMediaDBContent()

    def verifyTag(self, filePath, inlink_path=None):
        (inDBfile, realFile) = _TestDeejayDBLibrary.verifyTag(self, filePath,
                                                              inlink_path)

        for tag in ("title","artist","album","genre"):
            self.assert_(realFile[tag] == inDBfile[tag],
                "tag %s for %s different between DB and reality %s != %s" % \
                (tag,realFile["filename"],realFile[tag],inDBfile[tag]))

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# This work is based on the perl Video::Xine API.
# http://search.cpan.org/~stephen/Video-Xine/


import ctypes, os
import locale
import _xinelib as xinelib
import x11


class XineError(Exception): pass


class Osd(object):

    XINE_TEXT_PALETTE_SIZE = 11

    XINE_OSD_TEXT1  = (0 * XINE_TEXT_PALETTE_SIZE)
    XINE_OSD_TEXT2  = (1 * XINE_TEXT_PALETTE_SIZE)
    XINE_OSD_TEXT3  = (2 * XINE_TEXT_PALETTE_SIZE)
    XINE_OSD_TEXT4  = (3 * XINE_TEXT_PALETTE_SIZE)
    XINE_OSD_TEXT5  = (4 * XINE_TEXT_PALETTE_SIZE)
    XINE_OSD_TEXT6  = (5 * XINE_TEXT_PALETTE_SIZE)
    XINE_OSD_TEXT7  = (6 * XINE_TEXT_PALETTE_SIZE)
    XINE_OSD_TEXT8  = (7 * XINE_TEXT_PALETTE_SIZE)
    XINE_OSD_TEXT9  = (8 * XINE_TEXT_PALETTE_SIZE)
    XINE_OSD_TEXT10 = (9 * XINE_TEXT_PALETTE_SIZE)

    # white text, black border, transparent background
    XINE_TEXTPALETTE_WHITE_BLACK_TRANSPARENT    = 0
    # white text, noborder, transparent background
    XINE_TEXTPALETTE_WHITE_NONE_TRANSPARENT     = 1
    # white text, no border, translucid background
    XINE_TEXTPALETTE_WHITE_NONE_TRANSLUCID      = 2
    # yellow text, black border, transparent background
    XINE_TEXTPALETTE_YELLOW_BLACK_TRANSPARENT   = 3

    XINE_OSD_CAP_FREETYPE2 = 0x0001
    XINE_OSD_CAP_UNSCALED  = 0x0002

    def __init__(self, stream, width, height):
        self.__osd_p = xinelib.xine_osd_new(stream.stream_p(),
                                            0, 0, width, height)
        self.color_base = Osd.XINE_OSD_TEXT1
        self.__last_text = None

    def set_font(self, font_name, font_size):
        xinelib.xine_osd_set_font(self.__osd_p, font_name, font_size)

    def set_text_palette(self, palette_number, color_base):
        self.color_base = color_base
        xinelib.xine_osd_set_text_palette(self.__osd_p,
                                          palette_number,
                                          color_base)

    def is_unscaled(self):
        unscaled = xinelib.xine_osd_get_capabilities(self.__osd_p)\
                   & Osd.XINE_OSD_CAP_UNSCALED
        return unscaled and True or False

    def clear(self):
        xinelib.xine_osd_clear(self.__osd_p)

    def draw_text(self, posx, posy, text):
        xinelib.xine_osd_draw_text(self.__osd_p,
                                   0, 0, text, self.color_base)
        xinelib.xine_osd_set_position(self.__osd_p, posx, posy)
        self.__last_text = text

    def show(self):
        if self.is_unscaled:
            xinelib.xine_osd_show_unscaled(self.__osd_p, 0)
        else:
            xinelib.xine_osd_show(self.__osd_p, 0)

    def hide(self, text):
        if text == self.__last_text:
            xinelib.xine_osd_hide(self.__osd_p, 0)

    def close(self):
        self.clear()
        xinelib.xine_osd_free(self.__osd_p)
        self.__osd_p = None


class Event(object):

    XINE_EVENT_UI_PLAYBACK_FINISHED = 1
    XINE_EVENT_UI_CHANNELS_CHANGED  = 2
    XINE_EVENT_UI_SET_TITLE         = 3
    XINE_EVENT_UI_MESSAGE           = 4
    XINE_EVENT_FRAME_FORMAT_CHANGE  = 5
    XINE_EVENT_AUDIO_LEVEL          = 6
    XINE_EVENT_QUIT                 = 7
    XINE_EVENT_PROGRESS             = 8

    def __init__(self, type, contents):
        self.type = type
        if self.type == Event.XINE_EVENT_UI_MESSAGE:
            self.data = ctypes.cast(contents.data,
                             ctypes.POINTER(xinelib.xine_ui_message_data_t))
        else:
            self.data = None

    def message(self):
        if not self.data: return None
        msg = self.data.contents
        if msg.type != XinePlayer.XINE_MSG_NO_ERROR:
            if msg.explanation:
                message_txt = ctypes.string_at(ctypes.addressof(msg)\
                                               + msg.explanation)
                message_parameters = []
                param_address = ctypes.addressof(msg) + msg.parameters
                for param_index in range(0, msg.num_parameters):
                    message_par = ctypes.string_at(param_address)
                    param_address += len(message_par) + 1 # Skip '\0'
                    message_parameters.append(message_par)
                message_params = ' '.join(message_parameters)
                message = "%s %s" % (message_txt, message_params)
            else:
                raise XineError(msg.type)
        else:
            message = None
        return message.decode(locale.getpreferredencoding())


class EventQueue(object):

    def __init__(self, stream):
        self.__callbacks = []
        self.__event_queue_p = xinelib.xine_event_new_queue(stream.stream_p())

    def add_callback(self, callback, user_data=None):
        cb = xinelib.xine_event_listener_cb_t(self.__get_cb(callback))
        self.__callbacks.append(cb)
        xinelib.xine_event_create_listener_thread(self.__event_queue_p,
                                                  cb, user_data)

    def __get_cb(self, callback):
        def c_cb(user_data, event):
            callback(user_data, Event(event.contents.type, event.contents))
        return c_cb

    def close(self):
        xinelib.xine_event_dispose_queue(self.__event_queue_p)
        self.__callbacks = []


class AudioDriver(object):

    def __init__(self, xine, id=None, data=None):
        self.__xine_p = xine.xine_p()
        self.__driver_p = xinelib.xine_open_audio_driver(self.__xine_p,
                                                         id, data)
        if not self.__driver_p:
            raise XineError('Could not open audio driver')

    def driver_p(self):
        return self.__driver_p

    def destroy(self):
        xinelib.xine_close_audio_driver(self.__xine_p, self.__driver_p)
        self.__driver_p = None


class VideoDriver(object):

    XINE_VISUAL_TYPE_NONE    = 0
    XINE_VISUAL_TYPE_X11     = 1
    XINE_VISUAL_TYPE_X11_2   = 10
    XINE_VISUAL_TYPE_AA      = 2
    XINE_VISUAL_TYPE_FB      = 3
    XINE_VISUAL_TYPE_GTK     = 4
    XINE_VISUAL_TYPE_DFB     = 5
    XINE_VISUAL_TYPE_PM      = 6
    XINE_VISUAL_TYPE_DIRECTX = 7
    XINE_VISUAL_TYPE_CACA    = 8
    XINE_VISUAL_TYPE_MACOSX  = 9
    XINE_VISUAL_TYPE_XCB     = 11

    def __init__(self, xine, id=None, display_id=':0.0', fullscreen=False):
        self.__xine_p = xine.xine_p()

        try:
            visual = self.__make_x11_visual(display_id, fullscreen)
        except x11.X11Error:
            raise XineError('Could not initialize Xine on %s', display_id)

        self.__driver_p = xinelib.xine_open_video_driver(self.__xine_p,
                           id,
                           VideoDriver.XINE_VISUAL_TYPE_X11,
                           ctypes.cast(ctypes.byref(visual), ctypes.c_void_p))

        if not self.__driver_p:
            raise XineError('Could not open video driver')

    def driver_p(self):
        return self.__driver_p

    def __make_x11_visual(self, display_id, fullscreen=False):
        self.display = x11.X11Display(display_id)
        if fullscreen:
            self.window = self.display.do_create_window(fullscreen=True)
        else:
            self.window = self.display.do_create_window(320, 200)

        # Those callbacks are required to be kept in this tuple in order to
        # be safe from the garbage collector.
        self.__x11_callbacks = ( xinelib.xine_dest_size_cb(self.__dest_size_cb),
                           xinelib.xine_frame_output_cb(self.__frame_output_cb),
                               )

        vis = xinelib.x11_visual_t()
        vis.display = self.display.display_p()
        vis.screen = self.display.get_default_screen_number()
        vis.d = self.window.window_p()
        vis.frame_output_cb = ctypes.cast(self.__x11_callbacks[1],
                                          ctypes.c_void_p)
        vis.dest_size_cb = ctypes.cast(self.__x11_callbacks[0], ctypes.c_void_p)
        return vis

    def __frame_output_cb(self, user_data,
                                video_width, video_height, video_pixel_aspect,
                                dest_x, dest_y,
                                dest_width, dest_height, dest_pixel_aspect,
                                win_x, win_y):
        dest_x[0] = 0
        dest_y[0] = 0
        win_x[0] = self.window.video_area_info['win_x']
        win_y[0] = self.window.video_area_info['win_y']
        self.__dest_size_cb(user_data,
                            video_width, video_height, video_pixel_aspect,
                            dest_width, dest_height, dest_pixel_aspect)

    def __dest_size_cb(self, user_data,
                             video_width, video_height, video_pixel_aspect,
                             dest_width, dest_height, dest_pixel_aspect):
        dest_width[0] = self.window.video_area_info['width']
        dest_height[0] = self.window.video_area_info['height']
        dest_pixel_aspect[0] = self.window.video_area_info['aspect']

    XINE_GUI_SEND_DRAWABLE_CHANGED       = 2
    XINE_GUI_SEND_EXPOSE_EVENT           = 3
    XINE_GUI_SEND_TRANSLATE_GUI_TO_VIDEO = 4
    XINE_GUI_SEND_VIDEOWIN_VISIBLE       = 5
    XINE_GUI_SEND_SELECT_VISUAL          = 8
    XINE_GUI_SEND_WILL_DESTROY_DRAWABLE  = 9

    def send_gui_data(self, type, data=None):
        xinelib.xine_port_send_gui_data(self.__driver_p, ctypes.c_int(type),
                                        ctypes.cast(data, ctypes.c_void_p))

    def destroy(self):
        xinelib.xine_close_video_driver(self.__xine_p, self.__driver_p)
        self.__x11_callbacks = None
        self.__driver = None

        self.window.close()
        self.window = None
        self.display.destroy()
        self.display = None


class Stream(object):

    XINE_SPEED_PAUSE                 =  0
    XINE_SPEED_SLOW_4                =  1
    XINE_SPEED_SLOW_2                =  2
    XINE_SPEED_NORMAL                =  4
    XINE_SPEED_FAST_2                =  8
    XINE_SPEED_FAST_4                = 16

    def __init__(self, xine, audio_port=None, video_port=None):
        self.__xine = xine
        self.__event_queue = None
        self.__software_mixer = False
        self.__audio_port = audio_port or AudioDriver(xine)
        self.__video_port = video_port
        if self.__video_port:
            video_driver_p = self.__video_port.driver_p()
        else:
            video_driver_p = None
        self.__stream_p = xinelib.xine_stream_new(self.__xine.xine_p(),
                                                  self.__audio_port.driver_p(),
                                                  video_driver_p)

        if self.__xine.has_gapless():
            self.set_param(Stream.XINE_PARAM_EARLY_FINISHED_EVENT, 1)

        if self.__video_port:
            self.__video_port.send_gui_data(\
                              VideoDriver.XINE_GUI_SEND_DRAWABLE_CHANGED,
                              self.__video_port.window.window_p())
            self.__video_port.send_gui_data(\
                              VideoDriver.XINE_GUI_SEND_VIDEOWIN_VISIBLE,
                              1)
        else:
            self.set_param(Stream.XINE_PARAM_IGNORE_VIDEO, 1)
            self.set_param(Stream.XINE_PARAM_IGNORE_SPU, 1)

        self.__osd = None

    def has_video(self):
        if self.__video_port:
            return True
        else:
            return False

    def stream_p(self):
        return self.__stream_p

    def add_event_callback(self, callback):
        if not self.__event_queue:
            self.__event_queue = EventQueue(self)
        self.__event_queue.add_callback(callback)

    def open(self, mrl):
        if not xinelib.xine_open(self.__stream_p, mrl):
            raise XineError('Could not open %s' % mrl)

    def play(self, start_pos=0, start_time=0):
        if not xinelib.xine_play(self.__stream_p, ctypes.c_int(start_pos),
                                                  ctypes.c_int(start_time)):
            raise XineError('Could not play stream')
        else:
            self.set_dpms(False)

    def stop(self):
        self.set_dpms(False)
        xinelib.xine_stop(self.__stream_p)

    def get_pos_length(self):
        _pos_stream = ctypes.c_int()
        _pos_time = ctypes.c_int()
        _length_time = ctypes.c_int()
        result = xinelib.xine_get_pos_length(self.__stream_p,
                                             ctypes.byref(_pos_stream),
                                             ctypes.byref(_pos_time),
                                             ctypes.byref(_length_time))
        if result:
            return _pos_stream.value, _pos_time.value, _length_time.value
        else:
            return 0, 0, 0

    def get_pos(self):
        # Workaround for problems when you seek too quickly
        i = 0
        while i < 4:
            pos_s, pos_t, length = self.get_pos_length()
            if int(pos_t) > 0:  break
            xinelib.xine_usec_sleep(100000)
            i += 1
        return int(pos_t / 1000)

    def get_length(self):
        pos_s, pos_t, length = self.get_pos_length()
        return length / 1000

    XINE_STATUS_IDLE = 0
    XINE_STATUS_STOP = 1
    XINE_STATUS_PLAY = 2
    XINE_STATUS_QUIT = 3

    def get_status(self):
        return xinelib.xine_get_status(self.__stream_p)

    def set_fullscreen(self, fullscreen):
        # FIXME : This does not work yet.
        raise NotImplementedError
        if not self.__video_port:
            raise XineError('Stream is audio only')
        self.__video_port.window.set_fullscreen(fullscreen)
        self.__video_port.send_gui_data(\
                              VideoDriver.XINE_GUI_SEND_DRAWABLE_CHANGED,
                              self.__video_port.window.window_p())

    XINE_PARAM_SPEED                 =  1 # see below
    XINE_PARAM_AV_OFFSET             =  2 # unit: 1/90000 sec
    XINE_PARAM_AUDIO_CHANNEL_LOGICAL =  3 # 1 => auto, -2 => off
    XINE_PARAM_SPU_CHANNEL           =  4
    XINE_PARAM_VIDEO_CHANNEL         =  5
    XINE_PARAM_AUDIO_VOLUME          =  6 # 0..100
    XINE_PARAM_AUDIO_MUTE            =  7 # 1=>mute, 0=>unmute
    XINE_PARAM_AUDIO_COMPR_LEVEL     =  8 # <100=>off, % compress otherw
    XINE_PARAM_AUDIO_AMP_LEVEL       =  9 # 0..200, 100=>100% (default)
    XINE_PARAM_AUDIO_REPORT_LEVEL    = 10 # 1=>send events, 0=> don't
    XINE_PARAM_VERBOSITY             = 11 # control console output
    XINE_PARAM_SPU_OFFSET            = 12 # unit: 1/90000 sec
    XINE_PARAM_IGNORE_VIDEO          = 13 # disable video decoding
    XINE_PARAM_IGNORE_AUDIO          = 14 # disable audio decoding
    XINE_PARAM_IGNORE_SPU            = 15 # disable spu decoding
    XINE_PARAM_BROADCASTER_PORT      = 16 # 0: disable, x: server port
    XINE_PARAM_METRONOM_PREBUFFER    = 17 # unit: 1/90000 sec
    XINE_PARAM_EQ_30HZ               = 18 # equalizer gains -100..100
    XINE_PARAM_EQ_60HZ               = 19 # equalizer gains -100..100
    XINE_PARAM_EQ_125HZ              = 20 # equalizer gains -100..100
    XINE_PARAM_EQ_250HZ              = 21 # equalizer gains -100..100
    XINE_PARAM_EQ_500HZ              = 22 # equalizer gains -100..100
    XINE_PARAM_EQ_1000HZ             = 23 # equalizer gains -100..100
    XINE_PARAM_EQ_2000HZ             = 24 # equalizer gains -100..100
    XINE_PARAM_EQ_4000HZ             = 25 # equalizer gains -100..100
    XINE_PARAM_EQ_8000HZ             = 26 # equalizer gains -100..100
    XINE_PARAM_EQ_16000HZ            = 27 # equalizer gains -100..100
    XINE_PARAM_AUDIO_CLOSE_DEVICE    = 28 # force closing audio device
    XINE_PARAM_AUDIO_AMP_MUTE        = 29 # 1=>mute, 0=>unmute
    XINE_PARAM_FINE_SPEED            = 30 # 1.000.000 => normal speed
    XINE_PARAM_EARLY_FINISHED_EVENT  = 31 # send event when demux finish
    XINE_PARAM_GAPLESS_SWITCH        = 32 # next stream only gapless swi
    XINE_PARAM_DELAY_FINISHED_EVENT  = 33 # 1/10sec,0=>disable,-1=>f

    XINE_PARAM_VO_DEINTERLACE        = 0x01000000 # bool
    XINE_PARAM_VO_ASPECT_RATIO       = 0x01000001 # see below
    XINE_PARAM_VO_HUE                = 0x01000002 # 0..65535
    XINE_PARAM_VO_SATURATION         = 0x01000003 # 0..65535
    XINE_PARAM_VO_CONTRAST           = 0x01000004 # 0..65535
    XINE_PARAM_VO_BRIGHTNESS         = 0x01000005 # 0..65535
    XINE_PARAM_VO_ZOOM_X             = 0x01000008 # percent
    XINE_PARAM_VO_ZOOM_Y             = 0x0100000d # percent
    XINE_PARAM_VO_PAN_SCAN           = 0x01000009 # bool
    XINE_PARAM_VO_TVMODE             = 0x0100000a # ???
    XINE_PARAM_VO_WINDOW_WIDTH       = 0x0100000f # readonly
    XINE_PARAM_VO_WINDOW_HEIGHT      = 0x01000010 # readonly
    XINE_PARAM_VO_CROP_LEFT          = 0x01000020 # crop frame pixels
    XINE_PARAM_VO_CROP_RIGHT         = 0x01000021 # crop frame pixels
    XINE_PARAM_VO_CROP_TOP           = 0x01000022 # crop frame pixels
    XINE_PARAM_VO_CROP_BOTTOM        = 0x01000023 # crop frame pixels

    XINE_VO_ZOOM_STEP                = 100
    XINE_VO_ZOOM_MAX                 = 400
    XINE_VO_ZOOM_MIN                 = -85

    XINE_VO_ASPECT_AUTO              = 0
    XINE_VO_ASPECT_SQUARE            = 1 # 1:1
    XINE_VO_ASPECT_4_3               = 2 # 4:3
    XINE_VO_ASPECT_ANAMORPHIC        = 3 # 16:9
    XINE_VO_ASPECT_DVB               = 4 # 2.11:1
    XINE_VO_ASPECT_NUM_RATIOS        = 5

    def set_param(self, param, value):
        xinelib.xine_set_param(self.__stream_p, param, value)

    def get_param(self, param):
        return xinelib.xine_get_param(self.__stream_p, param)

    def set_software_mixer(self, software_mixer):
        self.__software_mixer = software_mixer

    def set_volume(self, volume):
        param = Stream.XINE_PARAM_AUDIO_VOLUME
        if self.__software_mixer:
            param = Stream.XINE_PARAM_AUDIO_AMP_LEVEL
        self.set_param(param, volume)

    XINE_META_INFO_TITLE             = 0
    XINE_META_INFO_COMMENT           = 1
    XINE_META_INFO_ARTIST            = 2
    XINE_META_INFO_GENRE             = 3
    XINE_META_INFO_ALBUM             = 4
    XINE_META_INFO_YEAR              = 5
    XINE_META_INFO_VIDEOCODEC        = 6
    XINE_META_INFO_AUDIOCODEC        = 7
    XINE_META_INFO_SYSTEMLAYER       = 8
    XINE_META_INFO_INPUT_PLUGIN      = 9

    def get_meta_info(self, info):
        return xinelib.xine_get_meta_info(self.__stream_p, info)

    XINE_STREAM_INFO_BITRATE         = 0
    XINE_STREAM_INFO_SEEKABLE        = 1
    XINE_STREAM_INFO_VIDEO_WIDTH     = 2
    XINE_STREAM_INFO_VIDEO_HEIGHT    = 3
    XINE_STREAM_INFO_VIDEO_RATIO     = 4

    def get_stream_info(self, info):
        return xinelib.xine_get_stream_info(self.__stream_p, info)

    XINE_LANG_MAX = 256

    def get_audio_lang(self, channel):
        _lang = ctypes.create_string_buffer(Stream.XINE_LANG_MAX)

        result = xinelib.xine_get_audio_lang(self.__stream_p, channel, _lang)
        if not result:
            _lang.raw = 'unknown'

        return _lang.value

    def get_spu_lang(self, channel):
        _lang = ctypes.create_string_buffer(Stream.XINE_LANG_MAX)

        result = xinelib.xine_get_spu_lang(self.__stream_p, channel, _lang)
        if not result:
            _lang.raw = 'unknown'

        return _lang.value

    def set_dpms(self, activated):
        if self.__video_port:
            self.__video_port.display.set_dpms(activated)

    def osd_new(self, font_size):
        if not self.__video_port:
            raise XineError('This Stream is audio only.')

        # Does this need locking the video area?
        self.__osd = Osd(self,
                         self.__video_port.window.video_area_info['width'],
                         self.__video_port.window.video_area_info['height'])
        self.__osd.set_font('sans', font_size)
        self.__osd.set_text_palette(\
                                   Osd.XINE_TEXTPALETTE_WHITE_BLACK_TRANSPARENT,
                                   Osd.XINE_OSD_TEXT1)
        return self.__osd

    def close(self):
        self.stop()
        xinelib.xine_close(self.__stream_p)

    def destroy(self):
        self.close()
        self.set_param(Stream.XINE_PARAM_AUDIO_CLOSE_DEVICE, 1)

        if self.__event_queue:
            self.__event_queue.close()
            self.__event_queue = None
        if self.__osd:
            self.__osd.close()
            self.__osd = None

        xinelib.xine_dispose(self.__stream_p)
        self.__stream_p = None

        self.__audio_port.destroy()
        self.__audio_port = None
        if self.__video_port:
            self.__video_port.destroy()
            self.__video_port = None


class XinePlayer(object):

    XINE_MSG_NO_ERROR              = 0 # (messages to UI)
    XINE_MSG_GENERAL_WARNING       = 1 # (warning message)
    XINE_MSG_UNKNOWN_HOST          = 2 # (host name)
    XINE_MSG_UNKNOWN_DEVICE        = 3 # (device name)
    XINE_MSG_NETWORK_UNREACHABLE   = 4 # none
    XINE_MSG_CONNECTION_REFUSED    = 5 # (host name)
    XINE_MSG_FILE_NOT_FOUND        = 6 # (file name or mrl)
    XINE_MSG_READ_ERROR            = 7 # (device/file/mrl)
    XINE_MSG_LIBRARY_LOAD_ERROR    = 8 # (library/decoder)
    XINE_MSG_ENCRYPTED_SOURCE      = 9 # none
    XINE_MSG_SECURITY              = 10 # (security message)
    XINE_MSG_AUDIO_OUT_UNAVAILABLE = 11 # none
    XINE_MSG_PERMISSION_ERROR      = 12 # (file name or mrl)
    XINE_MSG_FILE_EMPTY            = 13 # file is empty

    def __init__(self, config_file_path=None):
        self.__xine = xinelib.xine_new()
        if not self.__xine:
            raise XineError('Error during Xine instance initialisation')

        if config_file_path:
            xinelib.xine_config_load(self.__xine, config_file_path)
        else: # load default config file
            default = os.path.join(xinelib.xine_get_homedir(),".xine/config")
            xinelib.xine_config_load(self.__xine, default)

        xinelib.xine_init(self.__xine)

    def xine_p(self):
        return self.__xine

    def get_version(self):
        major = ctypes.c_int()
        minor = ctypes.c_int()
        sub = ctypes.c_int()
        xinelib.xine_get_version(ctypes.byref(major),
                                 ctypes.byref(minor),
                                 ctypes.byref(sub))
        return (major.value, minor.value, sub.value)

    def has_gapless(self):
        return xinelib.xine_check_version(1, 1, 1) == 1

    def get_supported_extensions(self):
        return xinelib.xine_get_file_extensions(self.__xine).split()

    def list_input_plugins(self):
        plugins = []
        for plugin in xinelib.xine_list_input_plugins(self.__xine):
            if not plugin:
                break
            plugins.append(plugin.lower())
        return plugins

    def destroy(self):
        xinelib.xine_exit(self.__xine)
        self.__xine = None

    XINE_ENGINE_PARAM_VERBOSITY_NONE = 0
    XINE_ENGINE_PARAM_VERBOSITY_LOG =  1

    def set_param(self, param, value):
        xinelib.xine_engine_set_param(self.__xine, param, value)

    def stream_new(self, audio_port=None, video_port=None, video=False):
        if video and not video_port:
            video_port = VideoDriver(self)
        return Stream(self, audio_port, video_port)


if __name__ == '__main__':
    import sys, os, time
    x = XinePlayer(os.path.expanduser('~/.xine/config'))
    print 'Xine %d.%d.%d ' % x.get_version()
    x.set_param(XinePlayer.XINE_ENGINE_PARAM_VERBOSITY_LOG, 1)
    vd = VideoDriver(x, fullscreen=False)
    s = x.stream_new(video_port=vd)

    def print_message(data, event):
        if event.type == Event.XINE_EVENT_UI_MESSAGE:
            try:
                print "Xine error message : %s" % event.message()
            except XineError, e:
                print "Xine exception message : %s " % e
    s.add_event_callback(print_message)

    file_path = sys.argv[1]
    if not file_path.startswith('/') and not file_path.startswith('http://'):
       file_path = os.path.join(os.getcwd(), file_path)
    if file_path.startswith('/'):
       file_path = 'file:/' + file_path
    s.open(file_path)
    s.play()
    time.sleep(15)
    s.stop()
    s.destroy()
    x.destroy()


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os, locale
import cPickle as pickle

from twisted.internet import reactor
from deejayd import mediafilters
from deejayd.interfaces import DeejaydError
from deejayd.component import SignalingComponent
from deejayd.mediadb.library import NotFoundException
from deejayd.sources._medialist import *
from deejayd.sources._playorder import orders


class SourceError(DeejaydError): pass


class _BaseSource(SignalingComponent):
    name = "unknown"
    _default_state = {"id": 1}

    def __init__(self, db):
        super(_BaseSource, self).__init__()
        self.db = db
        self._current = None
        self._playorder = orders["inorder"]

    def _load_state(self):
        state = self._default_state.copy()
        recorded_state = self.db.get_state("%s_state" % self.name)
        if recorded_state is not None:
            state = pickle.loads(recorded_state.encode("utf-8"))
        return state

    def get_recorded_id(self):
        return self._state["id"]

    def get_content(self, start = 0, stop = None):
        return self._media_list.get(start, stop)

    def get_current(self):
        return self._current

    def go_to(self, id, type = "id"):
        if type == "pos":
            try: id = self._media_list._order[id]
            except IndexError: return None
        self._current = self._playorder.set_explicit(self._media_list, id)
        return self._current

    def delete(self, ids):
        if not self._media_list.delete(ids):
            raise SourceError(_("Unable to delete selected ids"))
        self._media_list.reload_item_pos(self._current)

    def clear(self):
        self._media_list.clear()
        self._playorder.reset(self._media_list)

    def next(self, explicit = True):
        if explicit:
            self._current = self._playorder.next_explicit(self._media_list,\
                                                          self._current)
        else:
            self._current = self._playorder.next_implicit(self._media_list,\
                                                          self._current)
        return self._current

    def previous(self):
        self._current = self._playorder.previous_explicit(self._media_list,\
                                                          self._current)
        return self._current

    def get_status(self):
        return [
            (self.name, self._media_list.list_id),
            (self.name+"length", len(self._media_list)),
            (self.name+"timelength", self._media_list.time_length)
            ]

    def set_option(self, name, value):
        raise NotImplementedError

    def close(self):
        self._state["id"] = self._media_list.list_id
        states = [ (pickle.dumps(self._state), "%s_state" % self.name) ]
        self.db.set_state(states)


class _BaseLibrarySource(_BaseSource):
    available_playorder = ("inorder", "random", "onemedia","random-weighted")
    has_repeat = True
    _default_state = {"id": 1, "playorder": "inorder", "repeat": False}
    source_signal = ''

    def __init__(self, db, library):
        super(_BaseLibrarySource, self).__init__(db)
        self._state = self._load_state()
        if self.medialist_type == "sorted":
            self._media_list = SortedMediaList(self.get_recorded_id() + 1)
        elif self.medialist_type == "unsorted":
            self._media_list = UnsortedMediaList(self.get_recorded_id() + 1)
        self.library = library

        if self.has_repeat:
            self._media_list.repeat = self._state["repeat"]
        self._playorder = orders[self._state["playorder"]]

    def _get_playlist_content(self, pl_id):
        try:
            pl_id, name, type = self.db.is_medialist_exists(pl_id)
            if type == "static":
                return self.db.get_static_medialist(pl_id,\
                    infos=self.library.media_attr)
            elif type == "magic":
                properties = dict(self.db.get_magic_medialist_properties(pl_id))
                if properties["use-or-filter"] == "1":
                    filter = mediafilters.Or()
                else:
                    filter = mediafilters.And()
                sorts = mediafilters.DEFAULT_AUDIO_SORT
                if properties["use-limit"] == "1":
                    sorts = [(properties["limit-sort-value"],\
                             properties["limit-sort-direction"])] + sorts
                    limit = int(properties["limit-value"])
                else:
                    limit = None
                filter.filterlist = self.db.get_magic_medialist_filters(pl_id)
                return self.library.search(filter, sorts, limit)
        except TypeError:
            raise SourceError(_("Playlist %s does not exist.") % str(pl_id))

    def set_option(self, name, value):
        if name == "playorder":
            try: self._playorder = orders[value]
            except KeyError:
                raise SourceError(_("Unable to set %s order, not supported") %
                    value)
            else:
                self._state["playorder"] = value
        elif name == "repeat" and self.has_repeat:
            if not isinstance(value, bool):
                raise SourceError(_("Option value has to be a boolean"))
            self._media_list.repeat = value
            self._state["repeat"] = value
        else:
            raise NotImplementedError

    def get_status(self):
        status = super(_BaseLibrarySource, self).get_status()
        status.append((self.name+"playorder", self._playorder.name))
        if self.has_repeat:
            status.append((self.name+"repeat", self._media_list.repeat))
        return status

    def close(self):
        super(_BaseLibrarySource, self).close()
        self.db.set_static_medialist(self.base_medialist,self._media_list.get())

    def cb_library_changes(self, signal):
        file_id = signal.get_attr("id")
        getattr(self, "_%s_media" % signal.get_attr("type"))(file_id)

    def _add_media(self, media_id):
        pass

    def _update_media(self, media_id):
        try: media = self.library.get_file_withids([media_id])
        except NotFoundException:
            return
        if self._media_list.update_media(media[0]):
            self.dispatch_signame(self.source_signal)

    def _remove_media(self, media_id):
        if self._media_list.remove_media(media_id):
            self.dispatch_signame(self.source_signal)


class _BaseSortedLibSource(_BaseLibrarySource):
    medialist_type = "sorted"

    def set_sorts(self, sorts):
        for (tag, direction) in sorts:
            if tag not in self.sort_tags:
                raise SourceError(_("Tag '%s' not supported for sort") % tag)
            if direction not in ('ascending', 'descending'):
                raise SourceError(_("Bad sort direction for source"))
        self._sorts = sorts
        self._media_list.sort(self._sorts + self.default_sorts)
        self.dispatch_signame(self.source_signal)


class _BaseAudioLibSource(_BaseLibrarySource):
    base_medialist = ''
    medialist_type = "unsorted"

    def __init__(self, db, audio_library):
        super(_BaseAudioLibSource, self).__init__(db, audio_library)
        # load saved
        try: ml_id = self.db.get_medialist_id(self.base_medialist, 'static')
        except ValueError: # medialist does not exist
            pass
        else:
            self._media_list.set(self._get_playlist_content(ml_id))

    def add_song(self, song_ids, pos = None):
        try: medias = self.library.get_file_withids(song_ids)
        except NotFoundException:
            raise SourceError(_("One of these ids %s not found") % \
                ",".join(map(str, song_ids)))
        self._media_list.add_media(medias, pos)
        if pos: self._media_list.reload_item_pos(self._current)
        self.dispatch_signame(self.source_signal)

    def add_path(self, paths, pos = None):
        medias = []
        for path in paths:
            try: medias.extend(self.library.get_all_files(path))
            except NotFoundException:
                try: medias.extend(self.library.get_file(path))
                except NotFoundException:
                    raise SourceError(_("%s not found") % path)
        self._media_list.add_media(medias, pos)
        if pos: self._media_list.reload_item_pos(self._current)
        self.dispatch_signame(self.source_signal)

    def load_playlist(self, pl_ids, pos = None):
        medias = []
        for id in pl_ids:
            medias.extend(self._get_playlist_content(id))
        self._media_list.add_media(medias, pos)
        if pos: self._media_list.reload_item_pos(self._current)
        self.dispatch_signame(self.source_signal)

    def move(self, ids, new_pos):
        if not self._media_list.move(ids, new_pos):
            raise SourceError(_("Unable to move selected medias"))
        self._media_list.reload_item_pos(self._current)
        self.dispatch_signame(self.source_signal)

    def delete(self, ids):
        super(_BaseAudioLibSource, self).delete(ids)
        self.dispatch_signame(self.source_signal)

    def clear(self):
        self._current = None
        self._media_list.clear()
        self.dispatch_signame(self.__class__.source_signal)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import time, random
import threading

from deejayd.interfaces import DeejaydError
from deejayd.mediafilters import *


class InterfaceTests:
    """Test the deejayd daemon core interface, this test suite is to be used for testing the core facade and the client library."""

    def testSetMode(self):
        """Test setMode command"""

        # ask an unknown mode
        mode_name = self.testdata.getRandomString()
        ans = self.deejayd.set_mode(mode_name)
        self.assertRaises(DeejaydError, ans.get_contents)

        # ask a known mode
        known_mode = 'playlist'
        ans = self.deejayd.set_mode(known_mode)
        self.failUnless(ans.get_contents())

        # Test if the mode has been set
        status = self.deejayd.get_status().get_contents()
        self.assertEqual(status['mode'], known_mode)

    def testGetMode(self):
        """Test getMode command"""
        known_keys = ("playlist", "panel", "dvd", "webradio", "video")
        ans = self.deejayd.get_mode()
        for k in known_keys:
            self.failUnless(k in ans.get_contents().keys())
            self.failUnless(ans[k] in (True, False))

    def testGetStats(self):
        """Test getStats command"""
        ans = self.deejayd.get_stats()
        for k in ("audio_library_update","songs","artists","albums"):
            self.failUnless(k in ans.keys())

    def testPlaylistSaveRetrieve(self):
        """Save a playlist and try to retrieve it."""

        pl = []
        djplname = self.testdata.getRandomString()

        # Get current playlist
        djpl = self.deejayd.get_playlist()
        self.assertEqual(djpl.get().get_medias(), pl)

        ans = djpl.add_path(self.testdata.getRandomString())
        self.assertRaises(DeejaydError, ans.get_contents)

        # Add songs to playlist
        howManySongs = 3
        for songPath in self.test_audiodata.getRandomSongPaths(howManySongs):
            pl.append(self.test_audiodata.medias[songPath].tags["uri"])
            ans = djpl.add_path(songPath)
            self.failUnless(ans.get_contents())

        # Check for the playlist to be of appropriate length
        self.assertEqual(self.deejayd.get_status()['playlistlength'],
                         howManySongs)

        ans = djpl.save(djplname)

        self.failUnless(ans.get_contents())
        djpl_id = ans["playlist_id"]

        # Check for the saved playslit to be available
        retrievedPls = self.deejayd.get_playlist_list().get_medias()
        self.failUnless(djplname in [p["name"] for p in retrievedPls])

        # Retrieve the saved playlist
        djpl = self.deejayd.get_recorded_playlist(djpl_id, djplname, 'static')
        retrievedPl = djpl.get().get_medias()
        for song_nb in range(len(pl)):
            self.assertEqual(pl[song_nb], retrievedPl[song_nb]['uri'])

    def testPlaylistActions(self):
        """Test actions on current playlist."""
        djpl = self.deejayd.get_playlist()
        howManySongs = 4
        for songPath in self.test_audiodata.getRandomSongPaths(howManySongs):
            ans = djpl.add_path(songPath)
            self.failUnless(ans.get_contents())

        content = djpl.get().get_medias()
        song = self.testdata.getRandomElement(content)
        # shuffle the playlist
        ans = djpl.shuffle()
        self.failUnless(ans.get_contents())
        # move this song
        ans = djpl.move([song["id"]], 1)
        self.failUnless(ans.get_contents())
        # delete a song
        djpl.del_song(song["id"]).get_contents()
        content = djpl.get().get_medias()
        self.assertEqual(len(content), howManySongs-1)

    def testSavedStaticPlaylistActions(self):
        """Test action on saved static playlist"""
        djplname = self.testdata.getRandomString()
        djpl = self.deejayd.get_playlist()
        howManySongs = 3
        # Add songs to playlist
        for songPath in self.test_audiodata.getRandomSongPaths(howManySongs):
            ans = djpl.add_path(songPath)
            self.failUnless(ans.get_contents())
        # save playlist
        ans = djpl.save(djplname)
        self.failUnless(ans.get_contents())
        djpl_id = ans["playlist_id"]

        # add songs in the saved playlist
        savedpl = self.deejayd.get_recorded_playlist(djpl_id, djplname,\
                "static")
        for songPath in self.test_audiodata.getRandomSongPaths(howManySongs):
            # add twice the same song
            ans = savedpl.add_path(songPath)
            self.failUnless(ans.get_contents())
            ans = savedpl.add_path(songPath)
            self.failUnless(ans.get_contents())
        content = savedpl.get().get_medias()
        self.assertEqual(len(content), howManySongs*3)

    def testSavedMagicPlaylistActions(self):
        """Test action on saved magic playlist"""
        djplname = self.testdata.getRandomString()
        djpl_infos = self.deejayd.create_recorded_playlist(djplname,\
                "magic").get_contents()
        djpl = self.deejayd.get_recorded_playlist(djpl_infos["pl_id"],\
                djplname, "magic")
        filter = Equals('genre', self.test_audiodata.getRandomGenre())
        rnd_filter = Equals('genre', self.testdata.getRandomString())

        # add filter
        djpl.add_filter(filter).get_contents()
        # verify playlist
        ans = djpl.get()
        self.assertEqual(len(ans.get_filter()), 1)
        self.failUnless(len(ans.get_medias()) > 0)
        self.failUnless(filter.equals(ans.get_filter()[0]))

        # add random filter
        djpl.remove_filter(filter).get_contents()
        djpl.add_filter(rnd_filter).get_contents()
        # verify playlist
        ans = djpl.get()
        self.assertEqual(len(ans.get_filter()), 1)
        self.assertEqual(len(ans.get_medias()), 0)
        self.failUnless(rnd_filter.equals(ans.get_filter()[0]))

        # add filter and set property
        djpl.clear_filters().get_contents()
        djpl.add_filter(filter).get_contents()
        djpl.set_property("use-limit", "1").get_contents()
        djpl.set_property("limit-value", "1").get_contents()
        # verify playlist
        ans = djpl.get()
        self.assertEqual(len(ans.get_filter()), 1)
        self.assertEqual(len(ans.get_medias()), 1)
        self.failUnless(filter.equals(ans.get_filter()[0]))

    def testWebradioAddRetrieve(self):
        """Save a webradio and check it is in the list, then delete it."""

        wr_list = self.deejayd.get_webradios()

        # try to set wrong source
        ans = wr_list.set_source(self.testdata.getRandomString())
        self.assertRaises(DeejaydError, ans.get_contents)
        # set local source
        ans = wr_list.set_source('local')
        self.failUnless(ans.get_contents())
        # local does not have categorie, raise DeejaydError if we try to get it
        ans = wr_list.get_source_categories('local')
        self.assertRaises(DeejaydError, ans.get_contents)


        # Test for bad URI and inexistant playlist
        for badURI in [[self.testdata.getRandomString(50)],\
                   ['http://' + self.testdata.getRandomString(50) + '.pls']]:
            ans = wr_list.add_webradio(self.testdata.getRandomString(), badURI)
            self.assertRaises(DeejaydError, ans.get_contents)

        testWrName = self.testdata.getRandomString()

        testWrUrls = []
        for urlCount in range(self.testdata.getRandomInt(10)):
            testWrUrls.append('http://' + self.testdata.getRandomString(50))

        ans = wr_list.add_webradio(testWrName, testWrUrls)
        self.failUnless(ans.get_contents())

        wr_list = self.deejayd.get_webradios()
        wr_names = [wr['title'] for wr in wr_list.get().get_medias()]
        self.failUnless(testWrName in wr_names)

        retrievedWr = [wr for wr in wr_list.get().get_medias()\
                          if wr['title'] == testWrName][0]

        for url in testWrUrls:
            self.failUnless(url in retrievedWr['urls'])

        ans = wr_list.delete_webradio(51)
        self.assertRaises(DeejaydError, ans.get_contents)

        ans = wr_list.delete_webradio(retrievedWr['id'])
        self.failUnless(ans.get_contents())

        wr_list = self.deejayd.get_webradios().get().get_medias()
        self.failIf(testWrName in [wr['title'] for wr in wr_list])

    def testQueue(self):
        """Add songs to the queue, try to retrieve it, delete some songs in it, then clear it."""
        q = self.deejayd.get_queue()

        myq = []
        how_many_songs = 10
        for song_path in self.test_audiodata.getRandomSongPaths(how_many_songs):
            myq.append(self.test_audiodata.medias[song_path].tags["uri"])
            ans = q.add_path(song_path)
            self.failUnless(ans.get_contents())

        ddq = q.get()

        ddq_uris = [song['uri'] for song in ddq.get_medias()]
        for song_uri in myq:
            self.failUnless(song_uri in ddq_uris)

        random.seed(time.time())
        songs_to_delete = random.sample(myq, how_many_songs / 3)
        ans = q.del_songs([song['id'] for song in ddq.get_medias()\
                                      if song['uri'] in songs_to_delete])
        self.failUnless(ans.get_contents())

        ddq = q.get()
        ddq_uris = [song['uri'] for song in ddq.get_medias()]
        for song_uri in myq:
            if song_uri in songs_to_delete:
                self.failIf(song_uri in ddq_uris)
            else:
                self.failUnless(song_uri in ddq_uris)

        ans = q.clear()
        self.failUnless(ans.get_contents())
        ddq = q.get()
        self.assertEqual(ddq.get_medias(), [])

    def testPanel(self):
        """ Test panel source actions """
        panel = self.deejayd.get_panel()

        # get panel tags
        tags = panel.get_panel_tags().get_contents()
        for tag in tags:
            self.failUnless(tag in ['genre', 'artist',\
                    'various_artist', 'album'])

        # set filter
        bad_tag = self.testdata.getRandomString()
        random_str = self.testdata.getRandomString()
        ans = panel.set_panel_filters(bad_tag, [random_str]) # random tags
        self.assertRaises(DeejaydError, ans.get_contents)

        tag = self.testdata.getRandomElement(tags)
        panel.set_panel_filters(tag, [random_str]).get_contents()
        ans = panel.get()
        self.assertEqual([], ans.get_medias())

        # remove bad filter
        panel.remove_panel_filters(tag).get_contents()
        ans = panel.get()
        self.failUnless(len(ans.get_medias()) > 0)

        # get correct value for a tag
        result = self.deejayd.mediadb_list(tag, None).get_contents()
        value = self.testdata.getRandomElement(result)
        panel.set_panel_filters(tag, [value]).get_contents()
        ans = panel.get()
        self.failUnless(len(ans.get_medias()) > 0)

        # clear tag
        panel.clear_panel_filters().get_contents()
        ans = panel.get()
        self.failUnless(len(ans.get_medias()) > 0)

        # set search
        ans = panel.set_search_filter(bad_tag, random_str) # random tags
        self.assertRaises(DeejaydError, ans.get_contents)

        search_tag = self.testdata.getRandomElement(['title', 'album',\
                'genre', 'artist'])
        panel.set_search_filter(search_tag, random_str).get_contents()
        ans = panel.get()
        self.assertEqual([], ans.get_medias())

        # clear search
        panel.clear_search_filter().get_contents()
        ans = panel.get()
        self.failUnless(len(ans.get_medias()) > 0)

        # test sort
        result = panel.set_sorts([(bad_tag, 'ascending')])
        self.assertRaises(DeejaydError, result.get_contents)

        tag = self.testdata.getRandomElement(['title', 'rating', 'genre'])
        media_list = [m[tag] for m in ans.get_medias()]
        media_list.sort()
        panel.set_sorts([(tag, 'ascending')]).get_contents()
        ans = panel.get()
        for idx, m in enumerate(ans.get_medias()):
            self.assertEqual(m[tag], media_list[idx])

        # save a playlist and update active list
        djpl = self.deejayd.get_playlist()
        ans = self.deejayd.get_audio_dir()
        dir = self.testdata.getRandomElement(ans.get_directories())
        djpl.add_paths([dir]).get_contents()
        test_pl_name = self.testdata.getRandomString()
        djpl.save(test_pl_name).get_contents()
        pl_list = self.deejayd.get_playlist_list().get_medias()
        if len(pl_list) != 1:
            raise DeejaydError("playlist not saved")

        # save a playlist and update active list
        bad_plid = self.testdata.getRandomInt(2000, 1000)
        ans = panel.set_active_list("playlist", bad_plid)
        self.assertRaises(DeejaydError, ans.get_contents)

        panel.set_active_list("playlist", pl_list[0]["id"]).get_contents()
        panel_list = panel.get_active_list().get_contents()
        self.assertEqual(panel_list["type"], "playlist")
        self.assertEqual(panel_list["value"], pl_list[0]["id"])

    def testVideo(self):
        """ Test video source actions """
        if self.video_support:
            video_obj = self.deejayd.get_video()
            # choose a wrong directory
            rand_dir = self.testdata.getRandomString()
            ans = video_obj.set(rand_dir, "directory")
            self.assertRaises(DeejaydError, ans.get_contents)

            # get contents of root dir and try to set video directory
            ans = self.deejayd.get_video_dir()
            dir = self.testdata.getRandomElement(ans.get_directories())
            video_obj.set(dir, "directory").get_contents()

            # test videolist content
            video_list = video_obj.get().get_medias()
            ans = self.deejayd.get_video_dir(dir)
            self.assertEqual(len(video_list), len(ans.get_files()))
            # sort videolist content
            sort = [["rating", "ascending"]]
            video_obj.set_sorts(sort).get_contents()
            video_list = video_obj.get()
            self.assertEqual(video_list.get_sort(), sort)
            # set bad sort
            rnd_sort = [(self.testdata.getRandomString(), "ascending")]
            ans = video_obj.set_sorts(rnd_sort)
            self.assertRaises(DeejaydError, ans.get_contents)

            # search a wrong title
            rand = self.testdata.getRandomString()
            video_obj.set(rand, "search").get_contents()
            video_list = video_obj.get().get_medias()
            self.assertEqual(len(video_list), 0)
        else:
            try: video_obj = self.deejayd.get_video()
            except DeejaydError: # we test core
                pass
            else:
                ans = video_obj.get()
                self.assertRaises(DeejaydError, ans.get_medias)

    def testMediaRating(self):
        """Test media rating method"""
        # wrong media id
        random_id = self.testdata.getRandomInt(2000, 1000)
        ans = self.deejayd.set_media_rating([random_id], "2", "audio")
        self.assertRaises(DeejaydError, ans.get_contents)

        ans = self.deejayd.get_audio_dir()
        files = ans.get_files()
        file_ids = [f["media_id"] for f in files]
        # wrong rating
        ans = self.deejayd.set_media_rating(file_ids, "9", "audio")
        self.assertRaises(DeejaydError, ans.get_contents)
        # wrong library
        rand_lib = self.testdata.getRandomString()
        ans = self.deejayd.set_media_rating(file_ids, "2", rand_lib)
        self.assertRaises(DeejaydError, ans.get_contents)

        ans =  self.deejayd.set_media_rating(file_ids, "4", "audio")
        self.failUnless(ans.get_contents())
        ans = self.deejayd.get_audio_dir()
        files = ans.get_files()
        for f in files:
            self.assertEqual(4, int(f["rating"]))

    def testAudioLibrary(self):
        """ Test request on audio library (get_audio_dir, search)"""
        # try to get contents of an unknown directory
        rand_dir = self.testdata.getRandomString()
        ans = self.deejayd.get_audio_dir(rand_dir)
        self.assertRaises(DeejaydError, ans.get_contents)

        # get contents of root dir and try to get content of a directory
        ans = self.deejayd.get_audio_dir()
        dir = self.testdata.getRandomElement(ans.get_directories())
        ans = self.deejayd.get_audio_dir(dir)
        song_files = ans.get_files()
        self.failUnless(len(song_files) > 0)

        # search an unknown terms
        text = self.testdata.getRandomString()
        ans = self.deejayd.audio_search(text)
        self.assertEqual(ans.get_medias(), [])

        # search a known terms
        file = self.testdata.getRandomElement(song_files)
        ans = self.deejayd.audio_search(file["title"], "title")
        self.failUnless(len(ans.get_medias()) > 0)

    def testVideoLibrary(self):
        """ Test request on video library """
        if self.video_support:
            # try to get contents of an unknown directory
            rand_dir = self.testdata.getRandomString()
            ans = self.deejayd.get_video_dir(rand_dir)
            self.assertRaises(DeejaydError, ans.get_contents)

            # get contents of root dir and try to get content of a directory
            ans = self.deejayd.get_video_dir()
            dir = self.testdata.getRandomElement(ans.get_directories())
            ans = self.deejayd.get_video_dir(dir)
            files = ans.get_files()
            self.failUnless(len(files) > 0)
        else:
            ans = self.deejayd.get_video_dir()
            self.assertRaises(DeejaydError, ans.get_contents)

    def testSetOption(self):
        """ Test set_option commands"""
        # unknown option
        opt = self.testdata.getRandomString()
        ans = self.deejayd.set_option("playlist", opt, 1)
        self.assertRaises(DeejaydError, ans.get_contents)
        ans = self.deejayd.set_option(opt, "repeat", True)
        self.assertRaises(DeejaydError, ans.get_contents)

        ans = self.deejayd.set_option("webradio", "playorder", "inorder")
        self.assertRaises(DeejaydError, ans.get_contents)

        # set playlist option
        ans = self.deejayd.set_option("playlist", "playorder", "random")
        ans.get_contents()
        status = self.deejayd.get_status().get_contents()
        self.assertEqual(status["playlistplayorder"], "random")
        # set video option
        if self.video_support:
            ans = self.deejayd.set_option("video", "repeat", True)
            ans.get_contents()
            status = self.deejayd.get_status().get_contents()
            self.assertEqual(status["videorepeat"], True)

    def testAudioPlayer(self):
        """ Test player commands (play, pause,...) for audio """
        # try to set volume
        vol = 30
        ans = self.deejayd.set_volume(vol)
        self.failUnless(ans.get_contents())
        status = self.deejayd.get_status().get_contents()
        self.assertEqual(status["volume"], vol)

        # load songs in main playlist
        djpl = self.deejayd.get_playlist()
        ans = self.deejayd.get_audio_dir()
        dir = self.testdata.getRandomElement(ans.get_directories())
        djpl.add_paths([dir]).get_contents()

        # play song
        self.deejayd.set_mode("playlist").get_contents()
        self.deejayd.play_toggle().get_contents()
        # verify status
        status = self.deejayd.get_status().get_contents()
        self.assertEqual(status["state"], "play")
        # pause
        self.deejayd.play_toggle().get_contents()
        # verify status
        status = self.deejayd.get_status().get_contents()
        self.assertEqual(status["state"], "pause")
        self.deejayd.play_toggle().get_contents()
        # next and previous
        self.deejayd.next().get_contents()
        self.deejayd.previous().get_contents()
        status = self.deejayd.get_status().get_contents()
        self.assertEqual(status["state"], "play")
        # seek command
        self.deejayd.seek(1).get_contents()

        # test get_current command
        cur = self.deejayd.get_current().get_medias()
        self.assertEqual(len(cur), 1)

        self.deejayd.stop().get_contents()

    def testVideoPlayer(self):
        """ Test player commands (play, pause,...) for video """
        if self.video_support:
            video_obj = self.deejayd.get_video()
            # set video mode
            self.deejayd.set_mode("video").get_contents()

            # choose directory
            ans = self.deejayd.get_video_dir()
            dir = self.testdata.getRandomElement(ans.get_directories())
            video_obj.set(dir, "directory").get_contents()

            # play video file
            self.deejayd.play_toggle().get_contents()
            # verify status
            status = self.deejayd.get_status().get_contents()
            self.assertEqual(status["state"], "play")

            # test set_player_option cmd
            self.deejayd.set_player_option("av_offset", 100).get_contents()
            self.deejayd.set_player_option("aspect_ratio","16:9").get_contents()
            # bad option name
            ans = self.deejayd.set_player_option(\
                    self.testdata.getRandomString(),0)
            self.assertRaises(DeejaydError, ans.get_contents)
            # bad option value
            ans = self.deejayd.set_player_option("av_offset",\
                    self.testdata.getRandomString())
            self.assertRaises(DeejaydError, ans.get_contents)
            ans = self.deejayd.set_player_option("aspect_ratio",\
                    self.testdata.getRandomString())
            self.assertRaises(DeejaydError, ans.get_contents)

            self.deejayd.stop().get_contents()

    def testDvd(self):
        """ Test dvd commands"""
        if self.video_support:
            status = self.deejayd.get_status().get_contents()
            dvd_id = status["dvd"]

            self.deejayd.dvd_reload().get_contents()
            status = self.deejayd.get_status().get_contents()
            self.assertEqual(status["dvd"], dvd_id + 1)
        else:
            ans = self.deejayd.dvd_reload()
            self.assertRaises(DeejaydError, ans.get_contents)

    def test_mediadb_list(self):
        """Test db queries for tags listing."""

        tag = 'artist'
        filter = Or(Equals('genre', self.test_audiodata.getRandomGenre()),
                    Equals('genre', self.test_audiodata.getRandomGenre()))

        expected_tag_list = []

        for song_info in self.test_audiodata.medias.values():
            matches = False
            for f in filter.filterlist:
                if song_info.tags['genre'] == f.pattern:
                    matches = True
            if matches:
                if song_info.tags[tag] not in expected_tag_list:
                    expected_tag_list.append(song_info.tags[tag])

        result = self.deejayd.mediadb_list(tag, filter)

        for tagvalue in expected_tag_list:
            self.failUnless(tagvalue in result)


class InterfaceSubscribeTests:
    """Test the subscription interface. This is for the core and the async client only."""

    def test_subscription(self):
        """Checks that signals subscriptions get in and out."""
        server_notification = threading.Event()

        sub_id = self.deejayd.subscribe('player.status',
                                        lambda x: server_notification.set())
        self.failUnless((sub_id, 'player.status')\
                        in self.deejayd.get_subscriptions().items())

        self.deejayd.unsubscribe(sub_id)
        self.failUnless((sub_id, 'player.status')\
                        not in self.deejayd.get_subscriptions().items())

    def generic_sub_bcast_test(self, signal_name, trigger, trigger_args=()):
        """Checks that signal_name signal is broadcast when one of the trigger is involved."""
        server_notification = threading.Event()

        sub_id = self.deejayd.subscribe(signal_name,
                                        lambda x: server_notification.set())

        ans = trigger(*trigger_args)
        if ans:
            ans.get_contents()
        server_notification.wait(4)
        self.failUnless(server_notification.isSet(),
                        '%s signal was not broadcasted by %s.'\
                        % (signal_name, trigger.__name__))
        server_notification.clear()

        self.deejayd.unsubscribe(sub_id)

    def test_sub_broadcast_player_status(self):
        """Checks that player.status signals are broadcasted."""

        trigger_list = ((self.deejayd.play_toggle, ()),
                        (self.deejayd.set_option, ("playlist", 'repeat', 1)),
                        (self.deejayd.set_option, ('panel', "playorder",\
                                                   "random")),
                        (self.deejayd.set_volume, (51, )),
                        (self.deejayd.seek, (5, )),
                       )

        for trig in trigger_list:
            self.generic_sub_bcast_test('player.status', trig[0], trig[1])

    def test_sub_broadcast_player_current(self):
        """Checks that player.current signals are broadcasted."""

        trigger_list = ((self.deejayd.next, ()),
                        (self.deejayd.previous, ())
                       )

        for trig in trigger_list:
            self.generic_sub_bcast_test('player.current', trig[0], trig[1])

    def test_sub_broadcast_player_plupdate(self):
        """Checks that player.plupdate signals are broadcasted."""

        djpl = self.deejayd.get_playlist()
        ans = self.deejayd.get_audio_dir()
        dir = self.testdata.getRandomElement(ans.get_directories())

        trigger_list = ((djpl.add_paths, ([dir], )),
                        (djpl.shuffle, ()),
                        (djpl.clear, ()),
                       )

        for trig in trigger_list:
            self.generic_sub_bcast_test('player.plupdate', trig[0], trig[1])

    def test_sub_broadcast_playlist_listupdate(self):
        """Checks that playlist.listupdate signals are broadcasted."""

        djpl = self.deejayd.get_playlist()
        ans = self.deejayd.get_audio_dir()
        dir = self.testdata.getRandomElement(ans.get_directories())
        djpl.add_paths([dir]).get_contents()

        test_pl_name = self.testdata.getRandomString()
        test_pl_name2 = self.testdata.getRandomString()

        self.generic_sub_bcast_test('playlist.listupdate',
                                    djpl.save, (test_pl_name, ))

        retrievedPls = self.deejayd.get_playlist_list().get_medias()
        for pls in retrievedPls:
            if pls["name"] == test_pl_name:
                djpl_id = pls["id"]
                break
        trigger_list = (
                        (self.deejayd.erase_playlist, ([djpl_id], )),
                        (djpl.save, (test_pl_name2,)),
                       )

        for trig in trigger_list:
            self.generic_sub_bcast_test('playlist.listupdate', trig[0], trig[1])

    def test_sub_broadcast_playlist_update(self):
        """Checks that playlist.update signals are broadcasted."""
        ans = self.deejayd.get_audio_dir()
        dir = self.testdata.getRandomElement(ans.get_directories())
        filter = Equals('genre', self.test_audiodata.getRandomGenre())

        st_pl_name = self.testdata.getRandomString()
        mg_pl_name = self.testdata.getRandomString()

        st_pl_infos = self.deejayd.create_recorded_playlist(st_pl_name,\
                'static').get_contents()
        mg_pl_infos = self.deejayd.create_recorded_playlist(mg_pl_name,\
                'magic').get_contents()

        st_djpl = self.deejayd.get_recorded_playlist(st_pl_infos["pl_id"],\
                st_pl_name, 'static')
        mg_djpl = self.deejayd.get_recorded_playlist(mg_pl_infos["pl_id"],\
                mg_pl_name, 'magic')

        trigger_list = (
                        (st_djpl.add_path, (dir,)),
                        (mg_djpl.add_filter, (filter,)),
                        (mg_djpl.remove_filter, (filter,)),
                        (mg_djpl.set_property, ('use-limit', '1')),
                       )
        for trig in trigger_list:
            self.generic_sub_bcast_test('playlist.update', trig[0], trig[1])

    def test_sub_broadcast_panel_update(self):
        """Checks that panel.update signals are broadcasted."""

        # first save a playlist
        djpl = self.deejayd.get_playlist()
        ans = self.deejayd.get_audio_dir()
        dir = self.testdata.getRandomElement(ans.get_directories())
        djpl.add_paths([dir]).get_contents()
        test_pl_name = self.testdata.getRandomString()
        djpl.save(test_pl_name).get_contents()
        pl_list = self.deejayd.get_playlist_list().get_medias()
        if len(pl_list) != 1:
            raise DeejaydError("playlist not saved")

        djpn = self.deejayd.get_panel()

        trigger_list = (
                        (djpn.set_active_list, ("playlist", pl_list[0]["id"])),
                        (djpn.set_active_list, ("panel", "0")),
                        (djpn.set_panel_filters, ("genre", "zboub")),
                        (djpn.clear_panel_filters, []),
                        (djpn.set_sorts, ([("genre", "ascending")],)),
                       )

        for trig in trigger_list:
            self.generic_sub_bcast_test('panel.update', trig[0], trig[1])

    def test_sub_broadcast_video_update(self):
        """Checks that video.update signals are broadcasted."""

        if not self.video_support:
            return True

        djvideo = self.deejayd.get_video()
        ans = self.deejayd.get_video_dir()
        dir = self.testdata.getRandomElement(ans.get_directories())

        trigger_list = (
                        (djvideo.set, (dir, "directory")),
                        (djvideo.set_sorts, ([("title", "ascending")],)),
                       )

        for trig in trigger_list:
            self.generic_sub_bcast_test('video.update', trig[0], trig[1])

    def test_sub_broadcast_webradio_listupdate(self):
        """Checks that webradio.listupdate signals are broadcasted."""

        wr_list = self.deejayd.get_webradios()
        test_wr_name = self.testdata.getRandomString()
        test_wr_urls = 'http://' + self.testdata.getRandomString(50)

        self.generic_sub_bcast_test('webradio.listupdate',
                                    wr_list.add_webradio,
                                    (test_wr_name, test_wr_urls))

        retrieved_wr = [wr for wr in wr_list.get().get_medias()\
                           if wr['title'] == test_wr_name + '-1'][0]
        self.generic_sub_bcast_test('webradio.listupdate',
                                    wr_list.delete_webradio,
                                    (retrieved_wr['id'], ))

    def test_sub_broadcast_queue_update(self):
        """Checks that queue.update signals are broadcasted."""

        q = self.deejayd.get_queue()

        ans = self.deejayd.get_audio_dir()
        dir = self.testdata.getRandomElement(ans.get_directories())

        self.generic_sub_bcast_test('queue.update', q.add_paths, ([dir], ))

        retrieved_song_id = [song['id'] for song in q.get().get_medias()]
        self.generic_sub_bcast_test('queue.update',
                                    q.del_songs, (retrieved_song_id, ))

    def test_sub_broadcast_dvd_update(self):
        """Checks that dvd.update signals are broadcasted."""
        if not self.video_support:
            return True
        self.generic_sub_bcast_test('dvd.update', self.deejayd.dvd_reload, ())

    def test_sub_broadcast_mode(self):
        """Checks that mode signals are broadcasted."""
        self.generic_sub_bcast_test('mode', self.deejayd.set_mode, ('video', ))

    def test_sub_broadcast_mediadb_aupdate(self):
        """Checks that mediadb.aupdate signals are broadcasted."""

        # This is tested only using inotify support
        self.generic_sub_bcast_test('mediadb.aupdate',
                                    self.test_audiodata.addMedia)

    def test_sub_broadcast_mediadb_vupdate(self):
        """Checks that mediadb.vupdate signals are broadcasted."""
        if not self.video_support:
            return True
        # This is tested only using inotify support
        self.generic_sub_bcast_test('mediadb.vupdate',
                                    self.test_videodata.addMedia)


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from twisted.python import reflect
from deejayd import rpc as jsonrpclib

class JSONRPC(object):
    separator = '.'

    def __init__(self):
        self.subHandlers = {}

    def putSubHandler(self, prefix, handler):
        self.subHandlers[prefix] = handler

    def getSubHandler(self, prefix):
        return self.subHandlers.get(prefix, None)

    def getSubHandlerPrefixes(self):
        return self.subHandlers.keys()

    def _getFunction(self, functionPath):
        """Given a string, return a function, or raise jsonrpclib.Fault.

        This returned function will be called, and should return the result
        of the call, a Deferred, or a Fault instance.

        Override in subclasses if you want your own policy. The default
        policy is that given functionPath 'foo', return the method at
        self.jsonrpc_foo, i.e. getattr(self, "jsonrpc_" + functionPath).
        If functionPath contains self.separator, the sub-handler for
        the initial prefix is used to search for the remaining path.
        """
        if functionPath.find(self.separator) != -1:
            prefix, functionPath = functionPath.split(self.separator, 1)
            handler = self.getSubHandler(prefix)
            if handler is None:
                raise jsonrpclib.Fault(jsonrpclib.METHOD_NOT_FOUND,
                    "no such sub-handler %s" % prefix)
            return handler._getFunction(functionPath)

        f = getattr(self, "jsonrpc_%s" % functionPath, None)
        if not f:
            raise jsonrpclib.Fault(jsonrpclib.METHOD_NOT_FOUND,
                "function %s not found" % functionPath)
        elif not callable(f):
            raise jsonrpclib.Fault(jsonrpclib.METHOD_NOT_CALLABLE,
                "function %s not callable" % functionPath)
        else:
            return f

    def _listFunctions(self):
        """Return a list of the names of all jsonrpc methods."""
        return reflect.prefixedMethodNames(self.__class__, 'jsonrpc_')


class JSONRPCIntrospection(JSONRPC):
    """Implement a JSON-RPC Introspection API.

    By default, the methodHelp method returns the 'help' method attribute,
    if it exists, otherwise the __doc__ method attribute, if it exists,
    otherwise the empty string.

    To enable the methodSignature method, add a 'signature' method attribute
    containing a list of lists. See methodSignature's documentation for the
    format. Note the type strings should be JSON-RPC types, not Python types.
    """

    def __init__(self, parent):
        """Implement Introspection support for a JSONRPC server.

        @param parent: the JSONRPC server to add Introspection support to.
        """
        JSONRPC.__init__(self)
        self._jsonrpc_parent = parent

    def jsonrpc_listMethods(self):
        """Return a list of the method names implemented by this server."""
        functions = []
        todo = [(self._jsonrpc_parent, '')]
        while todo:
            obj, prefix = todo.pop(0)
            functions.extend([ prefix + name for name in obj._listFunctions() ])
            todo.extend([ (obj.getSubHandler(name),
                           prefix + name + obj.separator)
                          for name in obj.getSubHandlerPrefixes() ])
        return functions

    jsonrpc_listMethods.signature = [['array']]

    def jsonrpc_methodHelp(self, method):
        """
        Return a documentation string describing the use of the given method.
        """
        method = self._jsonrpc_parent._getFunction(method)
        return (getattr(method, 'help', None)
                or getattr(method, '__doc__', None) or '')

    jsonrpc_methodHelp.signature = [['string', 'string']]

    def jsonrpc_methodSignature(self, method):
        """Return a list of type signatures.

        Each type signature is a list of the form [rtype, type1, type2, ...]
        where rtype is the return type and typeN is the type of the Nth
        argument. If no signature information is available, the empty
        string is returned.
        """
        method = self._jsonrpc_parent._getFunction(method)
        return getattr(method, 'signature', None) or ''

    jsonrpc_methodSignature.signature = [['array', 'string'],
                                        ['string', 'string']]


def addIntrospection(jsonrpc):
    """Add Introspection support to an JSONRPC server.

    @param jsonrpc: The jsonrpc server to add Introspection support to.
    """
    jsonrpc.putSubHandler('system', JSONRPCIntrospection(jsonrpc))

# vim: ts=4 sw=4 expandtab
